from .base import Component
from tkinter import Canvas
import tkinter as tk

# Import modules needed for widget management
try:
    from .. import bounds_manager, defaults
except ImportError:
    import bounds_manager, defaults


class Container(Component):

    def __init__(
        self, root, width, height, fill=None, border=None, border_width=0, **kwargs
    ):
        self.initialized = False
        super().__init__(width, height)

        # Set up proper parent-child relationships for nebulatk widget hierarchy
        # The Container should be a child of the provided root, but act as a parent for its children
        self._root = root  # The Container's parent (usually the main window)
        self._window = (
            root if root.master == root else root.master
        )  # Reference to the top-level window

        # IMPORTANT: For containers, master should point to itself so that child widgets
        # draw on the container's canvas instead of the main window's canvas
        self.master = self

        # Container position (will be set when placed)
        self._container_x = 0
        self._container_y = 0

        # Add essential widget properties that the bounds manager and other systems expect
        self._orientation = 0  # Default orientation
        self.bounds_type = "default"  # Bounds type for hit detection
        self.state = False  # Whether the container is in an active state
        self.hovering = False  # Whether the container is being hovered
        self.visible = True
        self.can_focus = (
            True  # Allow container to receive focus so it can handle events
        )
        self.can_hover = True
        self.can_click = True

        # Add this container to its parent's children list
        self._root.children.append(self)

        # Initialize container-specific properties for managing child widgets
        self.children = []  # List to hold child widgets
        self.bounds = {}  # Bounds management for child widgets
        self.active = None  # Currently active (focused) widget
        self.down = None  # Widget currently being clicked
        self.hovered_child = None  # Child widget currently being hovered
        self.updates_all = (
            False  # Whether updates to members update the widget automatically
        )
        self.defaults = self._window.defaults  # Inherit defaults from window

        # Maps for canvas object replication
        self.maps = {}

        # Get the actual tkinter root window for creating the canvas
        # For nested containers, we need to traverse up to find the tkinter root
        current = self._window
        while hasattr(current, "_window") and current._window != current:
            current = current._window
        tkinter_root = current.root

        self.canvas = Canvas(tkinter_root, width=width, height=height)

        # Replicate existing canvas items from window
        for child in self._window.canvas.find_all():
            self.replicate_object(child)

        # Bind event handlers to the container's canvas
        self._bind_events()

        # Ensure proper layering after initialization
        self._ensure_proper_layering()

        self.initialized = True

    # Container acts as both a widget (has a root) and a parent (for its children)
    @property
    def root(self):
        """Return the Container's parent (what it's contained in)"""
        return self._root

    @root.setter
    def root(self, root):
        """Set the Container's parent"""
        if self._root is not None:
            self._root.children.remove(self)
        self._root = root
        self._window = root if root.master == root else root.master
        if root is not None:
            root.children.append(self)

    # Add the x and y properties that widgets expect from their parent
    @property
    def x(self):
        return self._container_x

    @x.setter
    def x(self, x):
        self._container_x = x

    @property
    def y(self):
        return self._container_y

    @y.setter
    def y(self, y):
        self._container_y = y

    # Add orientation property that bounds manager expects
    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        self._orientation = orientation

    # Add window property for accessing the original window_internal object
    # This is needed for image creation and other operations that need the top-level window
    @property
    def window(self):
        """Access the original window_internal object"""
        return self._window

    def _bind_events(self):
        """Bind event handlers for mouse and keyboard events"""
        self.canvas.bind("<Button-1>", self.click)
        self.canvas.bind("<ButtonRelease-1>", self.click_up)
        self.canvas.bind("<Motion>", self.hover)
        self.canvas.bind("<Leave>", self.leave_container)
        self.canvas.bind("<KeyPress>", self.typing)
        self.canvas.bind("<KeyRelease>", self.typing_up)
        self.canvas.focus_set()  # Allow canvas to receive keyboard events

    def click(self, event):
        """Handle mouse click events - similar to main window's click handler"""
        # Convert canvas coordinates to container-relative coordinates
        x = int(event.x)
        y = int(event.y)

        # Find the child widget that was clicked
        active_new = next(
            (child for child in self.children if bounds_manager.check_hit(child, x, y)),
            None,
        )

        # Update active widget
        if active_new is not self.active:
            if self.active is not None:
                self.active.change_active()
            self.active = active_new

        # Handle click down
        if active_new is not self.down:
            self.down = active_new
            if active_new is not None:
                active_new.clicked(x, y)

    def click_up(self, event):
        """Handle mouse click release events"""
        if self.down:
            self.down.release()
            self.down = None

    def hover(self, event):
        """Handle mouse hover events"""
        x = int(event.x)
        y = int(event.y)

        # Handle dragging if something is being clicked
        if self.down is not None:
            self.down.dragging(x, y)

        # Find hovered widget
        hovered_new = next(
            (child for child in self.children if bounds_manager.check_hit(child, x, y)),
            None,
        )

        # Update hovered widget
        if hovered_new is not self.hovered_child:
            if self.hovered_child is not None:
                self.hovered_child.hover_end()
            self.hovered_child = hovered_new
            if hovered_new is not None:
                hovered_new.hovered()

    def leave_container(self, event):
        """Handle mouse leaving container"""
        if self.hovered_child is not None:
            self.hovered_child.hover_end()
            self.hovered_child = None

    def typing(self, event):
        """Handle key press events"""
        if self.active is not None and self.active.can_type:
            self.active.typed(event.char)

    def typing_up(self, event):
        """Handle key release events"""
        # Can be extended for specific key release handling
        pass

    # Canvas drawing methods - delegate to container's canvas
    def create_image(self, x, y, image, state="normal"):
        """Create image on container's canvas"""
        # Get the tkinter-compatible image
        if hasattr(image, "tk_image"):
            # This is an image_manager.Image object
            tk_image = image.tk_image(self)
        else:
            # This is already a tkinter-compatible image
            tk_image = image

        image_id = self.canvas.create_image(
            x, y, image=tk_image, state=state, tags=("fg_item",)
        )
        return (image_id, image)

    def create_rectangle(
        self,
        x,
        y,
        width,
        height=0,
        fill=0,
        border_width=0,
        outline=None,
        state="normal",
    ):
        """Create rectangle on container's canvas"""
        if x == width or y == height:
            return None, None

        # Import image_manager here to avoid circular imports
        try:
            from .. import image_manager
        except ImportError:
            import image_manager

        # To support transparency with RGBA, check if the rectangle includes transparency
        if fill is not None and len(str(fill)) > 7 and str(fill)[7:] != "ff":
            bg_image = image_manager.create_image(
                fill, int(width - x), int(height - y), outline, border_width, self
            )
            id, image = self.create_image(x, y, bg_image, state=state)
            return id, image

        # Otherwise create a regular rectangle
        rect = self.canvas.create_rectangle(
            x,
            y,
            x + width,
            y + height,
            fill=fill[:7] if fill is not None else fill,
            outline=outline[:7] if outline is not None else outline,
            width=border_width,
            state=state,
            tags=("fg_item",),
        )
        return (rect, None)

    def create_text(
        self, x, y, text, font, fill="black", anchor="center", state="normal", angle=0
    ):
        """Create text on container's canvas"""
        text_id = self.canvas.create_text(
            x,
            y,
            text=text,
            font=font,
            fill=fill,
            anchor=anchor,
            state=state,
            angle=angle,
            tags=("fg_item",),
        )
        return (text_id, None)

    def move(self, _object, x, y):
        """Move object on container's canvas"""
        self.canvas.coords(_object, x, y)

    def object_place(self, _object, x, y):
        """Place object on container's canvas"""
        self.canvas.coords(_object, x, y)

    def delete(self, _object):
        """Delete object from container's canvas"""
        self.canvas.delete(_object)

    def change_state(self, _object, state):
        """Change state of object on container's canvas"""
        self.canvas.itemconfig(_object, state=state)

    def configure(self, _object=None, **kwargs):
        """Configure object on container's canvas"""
        if _object is not None:
            self.canvas.itemconfig(_object, **kwargs)

    def replicate_object(self, child):
        # Check if the object exists on the window canvas
        try:
            # Verify the object exists by checking if it's in the list of all items
            if child not in self._window.canvas.find_all():
                return  # Object doesn't exist, skip replication

            item_type = self._window.canvas.type(child)
            coords = self._window.canvas.coords(child)
            adjusted_coords = []
            for i, coord in enumerate(coords):
                if i % 2 == 0:  # x coordinates (even indices)
                    adjusted_coords.append(coord - self._container_x)
                else:  # y coordinates (odd indices)
                    adjusted_coords.append(coord - self._container_y)

            coords = adjusted_coords
        except tk.TclError:
            # Object doesn't exist or is invalid, skip replication
            return

        # During initialization, we replicate with original coordinates
        # Position adjustment will happen later in place() method
        if item_type == "image":
            self.maps[child] = self.canvas.create_image(
                coords[0],
                coords[1],
                image=self._window.canvas.itemcget(child, "image"),
                anchor=self._window.canvas.itemcget(child, "anchor"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "rectangle":
            self.maps[child] = self.canvas.create_rectangle(
                coords[0],
                coords[1],
                coords[2],
                coords[3],
                fill=self._window.canvas.itemcget(child, "fill"),
                outline=self._window.canvas.itemcget(child, "outline"),
                width=self._window.canvas.itemcget(child, "width"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "text":
            self.maps[child] = self.canvas.create_text(
                coords[0],
                coords[1],
                text=self._window.canvas.itemcget(child, "text"),
                font=self._window.canvas.itemcget(child, "font"),
                fill=self._window.canvas.itemcget(child, "fill"),
                anchor=self._window.canvas.itemcget(child, "anchor"),
                state=self._window.canvas.itemcget(child, "state"),
                angle=self._window.canvas.itemcget(child, "angle"),
                tags=("background_item",),
            )
        elif item_type == "line":
            self.maps[child] = self.canvas.create_line(
                *coords,
                fill=self._window.canvas.itemcget(child, "fill"),
                width=self._window.canvas.itemcget(child, "width"),
                stipple=self._window.canvas.itemcget(child, "stipple"),
                arrow=self._window.canvas.itemcget(child, "arrow"),
                arrowshape=self._window.canvas.itemcget(child, "arrowshape"),
                capstyle=self._window.canvas.itemcget(child, "capstyle"),
                joinstyle=self._window.canvas.itemcget(child, "joinstyle"),
                smooth=self._window.canvas.itemcget(child, "smooth"),
                splinesteps=self._window.canvas.itemcget(child, "splinesteps"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "oval":
            self.maps[child] = self.canvas.create_oval(
                coords[0],
                coords[1],
                coords[2],
                coords[3],
                fill=self._window.canvas.itemcget(child, "fill"),
                outline=self._window.canvas.itemcget(child, "outline"),
                width=self._window.canvas.itemcget(child, "width"),
                stipple=self._window.canvas.itemcget(child, "stipple"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "polygon":
            self.maps[child] = self.canvas.create_polygon(
                *coords,
                fill=self._window.canvas.itemcget(child, "fill"),
                outline=self._window.canvas.itemcget(child, "outline"),
                width=self._window.canvas.itemcget(child, "width"),
                stipple=self._window.canvas.itemcget(child, "stipple"),
                smooth=self._window.canvas.itemcget(child, "smooth"),
                splinesteps=self._window.canvas.itemcget(child, "splinesteps"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "arc":
            self.maps[child] = self.canvas.create_arc(
                coords[0],
                coords[1],
                coords[2],
                coords[3],
                fill=self._window.canvas.itemcget(child, "fill"),
                outline=self._window.canvas.itemcget(child, "outline"),
                width=self._window.canvas.itemcget(child, "width"),
                stipple=self._window.canvas.itemcget(child, "stipple"),
                start=self._window.canvas.itemcget(child, "start"),
                extent=self._window.canvas.itemcget(child, "extent"),
                style=self._window.canvas.itemcget(child, "style"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )
        elif item_type == "bitmap":
            self.maps[child] = self.canvas.create_bitmap(
                coords[0],
                coords[1],
                bitmap=self._window.canvas.itemcget(child, "bitmap"),
                anchor=self._window.canvas.itemcget(child, "anchor"),
                foreground=self._window.canvas.itemcget(child, "foreground"),
                background=self._window.canvas.itemcget(child, "background"),
                state=self._window.canvas.itemcget(child, "state"),
                tags=("background_item",),
            )

        # After replicating, ensure background items stay behind container widgets
        if self.canvas.find_withtag("fg_item"):
            self.canvas.tag_lower("background_item", "fg_item")
        else:
            self.canvas.tag_lower("background_item")

    def _show(self, root):
        pass

    def _hide(self, root):
        pass

    def place(self, x, y):
        """Place the container at specified position"""
        self._container_x = x
        self._container_y = y
        self.canvas.place(x=x, y=y)

        # Update positions of replicated background objects now that we know the container's position
        self._update_background_positions()

        return self

    def _update_background_positions(self):
        """Update the positions of replicated background objects relative to container position."""
        for original_item, replicated_item in self.maps.items():
            try:
                # Get original coordinates from the main window canvas
                if original_item not in self._window.canvas.find_all():
                    continue  # Original item no longer exists

                original_coords = self._window.canvas.coords(original_item)

                # Adjust coordinates relative to the container's position
                adjusted_coords = []
                for i, coord in enumerate(original_coords):
                    if i % 2 == 0:  # x coordinates (even indices)
                        adjusted_coords.append(coord - self._container_x)
                    else:  # y coordinates (odd indices)
                        adjusted_coords.append(coord - self._container_y)

                # Update the replicated object's position
                self.canvas.coords(replicated_item, *adjusted_coords)

            except tk.TclError:
                # Original item or replicated item no longer exists
                continue

    # Event handler methods that widgets are expected to have
    def hovered(self):
        """Handle hover events on the container"""
        self.hovering = True

    def hover_end(self):
        """Handle hover end events on the container"""
        self.hovering = False

    def clicked(self, x=None, y=None):
        """Handle click events on the container"""
        # Container can handle its own click logic here if needed
        pass

    def release(self):
        """Handle mouse release events on the container"""
        # Container can handle its own release logic here if needed
        pass

    def dragging(self, x, y):
        """Handle dragging events on the container"""
        # Container can handle its own dragging logic here if needed
        pass

    def change_active(self):
        """Handle active state changes"""
        # Container can handle its own active state changes here if needed
        pass

    def _ensure_proper_layering(self):
        """Ensure background items stay behind foreground items"""
        # Move all background items to the back
        if self.canvas.find_withtag("background_item"):
            self.canvas.tag_lower("background_item")
        # Move all foreground items to the front
        if self.canvas.find_withtag("fg_item"):
            self.canvas.tag_raise("fg_item")



# Import tkinter and filedialog
import sys
import threading
import tkinter as tk

# Import python standard packages
from time import sleep
from tkinter import filedialog as tkfile_dialog


# Importing from another file works differently than if running this file directly for some reason.
# Import all required modules from nebulatk
try:
    from . import (
        bounds_manager,
        fonts_manager,
        colors_manager,
        image_manager,
        standard_methods,
        defaults,
        animation_controller,
        taskbar_manager,
    )

    # Import Component and _widget classes from widgets.base
    from .widgets.base import Component, _widget, _widget_properties

    # Import widget classes from widgets module
    from .widgets import Button, Label, Entry, Frame, Slider, Container

except ImportError:
    import bounds_manager
    import fonts_manager
    import colors_manager
    import image_manager
    import standard_methods
    import defaults
    import animation_controller
    import taskbar_manager

    # Import Component and _widget classes from widgets.base
    from widgets.base import Component, _widget, _widget_properties

    # Import widget classes from widgets module
    from widgets import Button, Label, Entry, Frame, Slider, Container


# Our implementation of tkinter's file_dialog.
# This mostly just exists so that the user doesn't have to import nebulatk and tkinter
def FileDialog(window, initialdir=None, mode="r", filetypes=(("All files", "*"))):
    """Identical to tkinter.FileDialog

    Args:
        window (nebulatk.Window): Root window
        initialdir (str, optional): Initial directory to open to. Defaults to None.
        mode (str, optional): File open mode. Defaults to "r".
        filetypes (list, optional): List of filetypes. Format like:
            [
                ("Name","*.ext"),
                ("Name", ("*.ext", "*.ext2"))
            ]
            Defaults to [("All files", "*")].

    Returns:
        file: Open file
    """
    window.leave_window(None)
    file = tkfile_dialog.askopenfile(
        initialdir=initialdir,
        mode=mode,
        filetypes=filetypes,
    )
    window.leave_window(None)
    return file


# Internal window class to implement threading
# NOTE: Threading and tcl do not combine well, so we have to do a lot of stuff ourselves and be careful that all modifications to the tcl windows are done in the same thread
class _window_internal(threading.Thread, Component):

    def __init__(
        self,
        width=500,
        height=500,
        title="ntk",
        canvas_width="default",
        canvas_height="default",
        closing_command=None,
        resizable=(True, True),
        override=False,
    ):
        # Initialize the thread
        super().__init__()

        # Initialize the bounds
        # Bounds == structured as:
        # bounds[y] = [
        #   [_object, start_x, end_x],
        #   [_object2, start_x, end_x],
        # ]
        Component.__init__(self, width, height)
        self.bounds = {}

        # Initialize variables
        self.root = None
        self.master = self
        self.canvas = None
        self.down = None
        self.active = None
        self.hovered = None
        self.resizable = resizable
        self.override = override
        self.updates_all = (
            False  # Whether updates to members update the widget automatically
        )

        self.children = []
        self.active_animations = []

        self.active_keys = []

        self.title = title

        # Initialize canvas size
        if canvas_height == "default":
            canvas_height = self.height
        if canvas_width == "default":
            canvas_width = self.width

        self.canvas_width = canvas_width
        self.canvas_height = canvas_height

        # Because of threading and garbage collection, images created will sometimes have to be stored in this variable to ensure they do not get deleted
        self.images = {}

        # Initialize rest of variables
        self.running = True
        self.closing_command = closing_command

        self.defaults = defaults.new()

        self._taskbar_manager = None

    @property
    def taskbar_manager(self):
        if (
            self._taskbar_manager is None
            and sys.platform == "win32"
            and self.root is not None
        ):
            self._taskbar_manager = taskbar_manager.TaskbarManager(self)
        return self._taskbar_manager

    @taskbar_manager.setter
    def taskbar_manager(self, value):
        return "Cannot set taskbar_manager"

    # NOTE: EVENT HANDLERS

    # Handle mouse clicked down event
    def click(self, event):
        # Mouse position == a float, so convert to integer
        x = int(event.x)
        y = int(event.y)

        active_new = next(
            (child for child in self.children if bounds_manager.check_hit(child, x, y)),
            None,
        )
        """
        # Find the object that was clicked
        # Check if there are any objects initialized at that position
        if y in self.bounds:
            # Iterate over all objects at that y coordinate
            for i in self.bounds[y]:
                # Check if the mouse was within that object's boundary
                if x <= i[2] and x >= i[1]:
                    # Object found
                    # Set active_new and break
                    active_new = i[0]
                    break
        """
        # If the new object is actually new, update the current active widget
        # self.active is used for things like detecting what widget to send keypresses to
        if active_new is not self.active:
            if self.active is not None:
                self.active.change_active()
            self.active = active_new

        # If the new object wasn't already being clicked
        if active_new is not self.down:
            self.down = active_new

            # If the new object is a valid widget
            if active_new is not None:
                # Click on the object with click position
                active_new.clicked(x, y)

    # Handle mouse click up events
    def click_up(self, event):
        # If something was being clicked on, release the widget
        if self.down:
            self.down.release()
            self.down = None

    # Handle mouse movement event
    def hover(self, event):
        # Mouse position == a float, so convert to integer
        x = int(event.x)
        y = int(event.y)

        # Temp variable hovered_new
        # Default to None in case no object was hovered over

        # If something is already being clicked on, this is also a dragging event on that widget.
        if self.down is not None:
            self.down.dragging(x, y)

        hovered_new = next(
            (child for child in self.children if bounds_manager.check_hit(child, x, y)),
            None,
        )

        if hovered_new is not self.hovered and hovered_new is not None:
            hovered_new.hovered()
        """
        # Find the object that was hovered over
        # Check if there are any objects initialized at that position
        if y in self.bounds:
            # Iterate over all objects at that y coordinate
            for i in self.bounds[y]:
                # Check if the mouse was within that object's boundary
                if x <= i[2] and x >= i[1]:
                    # Object found
                    # If this object wasn't already being hovered over, trigger new object's hovered event
                    if i[0] is not self.hovered:
                        print("hovered")
                        i[0].hovered()

                    # Set hovered_new and break
                    hovered_new = i[0]
                    break
        """
        # If the object wasn't already being hovered over, check that there was something being hovered over previously and trigger the old object's hover_end event
        if hovered_new is not self.hovered:
            if self.hovered is not None:
                self.hovered.hover_end()

            # Update the currently hovered object
            self.hovered = hovered_new

    # Handle the mouse leaving the window
    def leave_window(self, event):
        # If something was being hovered, trigger the hover_end event
        if self.hovered is not None:
            self.hovered.hover_end()

        # If something was being clicked on, simulate the mouse button being released
        if self.down is not None:
            self.down.release()

        # Ensure that the variables for selected, clicked, and hovered objects are reset
        self.hovered = None
        self.down = None

    # Handle keystrokes
    def typing(self, event):
        # If a widget == active, we should send keystrokes to it
        if self.active:
            self.active.typed(event)

        if event.keysym not in self.active_keys:
            self.active_keys.append(event.keysym)

    def typing_up(self, event):
        if event.keysym in self.active_keys:
            self.active_keys.remove(event.keysym)

    # NOTE: CREATION HANDLERS
    # These are necessary so that threading works properly with tcl

    # Wrapper for canvas.create_image method
    def create_image(self, x, y, image, state="normal"):
        img = self.canvas.create_image(
            x,
            y,
            image=image.tk_image(self),
            anchor="nw",
            state=state,
        )
        for child_canvas in self.children:
            if type(child_canvas).__name__ == "Container" and child_canvas.initialized:
                child_canvas.replicate_object(img)
                # Ensure background items stay behind container widgets
                child_canvas.canvas.tag_lower("background_item")
                if child_canvas.canvas.find_withtag("fg_item"):
                    child_canvas.canvas.tag_raise("fg_item")
        return img, image

    # Wrapper for canvas.create_rectangle method
    def create_rectangle(
        self, x, y, widt, height=0, fill=0, border_width=0, outline=None, state="normal"
    ):
        if x == widt or y == height:
            return None, None
        # To support transparency with RGBA, we need to check whether the rectangle includes transparency
        if fill is not None and fill[7:] != "ff":
            bg_image = image_manager.create_image(
                fill, int(widt - x), int(height - y), outline, border_width, self
            )
            id, image = self.create_image(x, y, bg_image, state=state)
            return id, image

            # Otherwise we can continue with creating the rectangle

        rect = self.canvas.create_rectangle(
            x + border_width / 2,
            y + border_width / 2,
            widt - border_width / 2,
            height - border_width / 2,
            fill=fill[:7],
            width=border_width,
            outline=outline[:7],
            state=state,
        )

        for child_canvas in self.children:
            if type(child_canvas).__name__ == "Container" and child_canvas.initialized:
                child_canvas.replicate_object(rect)
                # Ensure background items stay behind container widgets
                child_canvas.canvas.tag_lower("background_item")
                if child_canvas.canvas.find_withtag("fg_item"):
                    child_canvas.canvas.tag_raise("fg_item")
        return (rect, None)

    # Wrapper for canvas.create_text method
    def create_text(
        self, x, y, text, font, fill="black", anchor="center", state="normal", angle=0
    ):

        text_id = self.canvas.create_text(
            x,
            y,
            text=text,
            font=font,
            fill=fill,
            anchor=anchor,
            state=state,
            angle=angle,
        )

        for child_canvas in self.children:
            if type(child_canvas).__name__ == "Container" and child_canvas.initialized:
                child_canvas.replicate_object(text_id)
                # Ensure background items stay behind container widgets
                child_canvas.canvas.tag_lower("background_item")
                if child_canvas.canvas.find_withtag("fg_item"):
                    child_canvas.canvas.tag_raise("fg_item")
        return (text_id, None)

    # Wrapper for canvas.move method
    def move(self, _object, x, y):
        if _object is not None:
            self.canvas.move(_object, x, y)
            for child_canvas in self.children:
                if (
                    type(child_canvas).__name__ == "Container"
                    and child_canvas.initialized
                ):
                    if _object in child_canvas.maps:
                        child_id = child_canvas.maps[_object]
                        child_canvas.canvas.move(child_id, x, y)
                    else:
                        # Only try to replicate if the object exists and is valid
                        try:
                            if _object in self.canvas.find_all():
                                child_canvas.replicate_object(_object)
                        except (tk.TclError, AttributeError):
                            # Object is invalid or doesn't exist, skip replication
                            pass

                        child_canvas.canvas.tag_lower("background_item")
                        if child_canvas.canvas.find_withtag("fg_item"):
                            child_canvas.canvas.tag_raise("fg_item")

    def object_place(self, _object, x, y):
        if _object is not None:
            coords = self.canvas.coords(_object)
            self.move(_object, x - coords[0], y - coords[1])

    # Wrapper for canvas.delete method
    def delete(self, _object):
        self.canvas.delete(_object)
        for child_canvas in self.children:
            if type(child_canvas).__name__ == "Container" and child_canvas.initialized:
                if _object in child_canvas.maps:
                    child_id = child_canvas.maps[_object]
                    child_canvas.canvas.delete(child_id)

    # Wrapper for canvas.change_state method
    def change_state(self, _object, state):
        self.canvas.itemconfigure(_object, state=state)
        for child_canvas in self.children:
            if type(child_canvas).__name__ == "Container" and child_canvas.initialized:
                if _object in child_canvas.maps:
                    child_id = child_canvas.maps[_object]
                    child_canvas.canvas.itemconfigure(child_id, state=state)
                else:
                    # Only try to replicate if the object exists and is valid
                    try:
                        if _object in self.canvas.find_all():
                            child_canvas.replicate_object(_object)
                    except (tk.TclError, AttributeError):
                        # Object is invalid or doesn't exist, skip replication
                        pass
                child_canvas.canvas.tag_lower("background_item")
                if child_canvas.canvas.find_withtag("fg_item"):
                    child_canvas.canvas.tag_raise("fg_item")

    # NOTE: Other methods

    # Handle window closing

    def close(self):
        # signal the poller → mainloop quits on next tick
        self.running = False

        # stop any running animations
        self.close_animations()

        self.root.after(200, self.root.quit)

        # wait up to a second for the thread to finish
        self.join(timeout=0.1)

        if self.closing_command:
            self.closing_command()

    def destroy(self):
        self.close()

    def close_animations(self):
        for anim in self.active_animations.copy():
            anim.stop()
            if anim.thread is not None:
                while anim.thread.is_alive():
                    self.root.update()
                    sleep(0.01)
                anim.thread.join(timeout=1)

    # Add in window.place() to simplify tcl's root.geometry method
    def place(self, x=0, y=0):
        self.root.geometry(f"{self.width}x{self.height}+{x}+{y}")
        return self

    # Wrapper for root.bind() method, in future, add global keypress handling
    def bind(self, key, command):
        self.root.bind(key, command)

    # Main method
    def run(self):
        # Create window
        self.root = tk.Tk()
        self.canvas = tk.Canvas(
            self.root,
            width=self.canvas_width,
            height=self.canvas_height,
            borderwidth=0,
            highlightthickness=0,
        )
        self.canvas.pack()

        # Initialize window
        self.root.geometry(f"{self.width}x{self.height}")

        self.root.title(self.title)

        self.root.resizable(self.resizable[0], self.resizable[1])

        self.root.overrideredirect(self.override)

        def close():
            self.running = False
            self.close_animations()
            try:
                self.root.quit()
            except Exception as e:
                print(f"exception{e}")
            if self.closing_command is not None:
                self.closing_command()

        # Bind events to our new event handlers
        self.bind("<Button-1>", self.click)
        self.bind("<ButtonRelease-1>", self.click_up)
        self.bind("<Key>", self.typing)
        self.bind("<KeyRelease>", self.typing_up)
        self.root.protocol("WM_DELETE_WINDOW", close)
        self.bind("<Motion>", self.hover)
        self.bind("<Leave>", self.leave_window)

        self.root.mainloop()

        # schedule a periodic check of `self.running`,
        # so that close() can break us out cleanly:
        # print("exited")
        # NOTE: The following code is an alternative, but broken, method of running mainloop
        """while self.running:
            self.root.update_idletasks()
            self.root.update()
            sleep(0.01)"""

    # Add resize method to change window dimensions
    def resize(self, width=None, height=None):
        """Resize the window to the specified dimensions.

        Args:
            width (int, optional): New width. If None, keeps current width.
            height (int, optional): New height. If None, keeps current height.

        Returns:
            self: Returns self for method chaining
        """
        if width is not None:
            self._size[0] = int(width)
        if height is not None:
            self._size[1] = int(height)

        # Update the window geometry
        self.root.geometry(f"{self.width}x{self.height}")

        # If canvas dimensions should match window, update those too
        self.canvas_width = self.width
        self.canvas.configure(width=self.width)
        self.canvas_height = self.height
        self.canvas.configure(height=self.height)

        # Update all children to match new dimensions
        # self._update_children()

        return self

    # Add show method similar to widget show
    def _show(self, root):
        """Show the window if it was previously hidden.

        Returns:
            self: Returns self for method chaining
        """
        # Use deiconify to make window visible if it was withdrawn
        if self.root is not None:
            self.root.deiconify()
            self.update()

        return self

    # Add hide method similar to widget hide
    def _hide(self, root):
        """Hide the window without destroying it.

        Returns:
            self: Returns self for method chaining
        """
        # Use withdraw to hide window without destroying it
        if self.root is not None:
            self.root.withdraw()

        return self

    # Add configure method similar to widget configure
    # Also Wrapper for canvas.configure method
    def configure(self, _object=None, **kwargs):
        """Configure window properties.

        Supported properties:
        - width: Window width
        - height: Window height
        - title: Window title
        - resizable: Tuple of (width_resizable, height_resizable)

        Returns:
            self: Returns self for method chaining
        """
        if _object:
            self.canvas.itemconfigure(_object, kwargs)
            return self

        if "width" in kwargs or "height" in kwargs:
            self.resize(kwargs.get("width"), kwargs.get("height"))

        if "title" in kwargs:
            self.title = kwargs["title"]
            if self.root is not None:
                self.root.title(self.title)

        if "resizable" in kwargs:
            self.resizable = kwargs["resizable"]
            if type(self.resizable) is bool:
                self.resizable = (self.resizable, self.resizable)
            if self.root is not None:
                self.root.resizable(self.resizable[0], self.resizable[1])

        return self

    def update(self):
        self.root.update()
        return self


def Window(
    width=500,
    height=500,
    title=None,
    canvas_width="default",
    canvas_height="default",
    closing_command=None,
    resizable=(True, True),
    override=False,
):
    """Window constructor

    Args:
        width (int, optional): Width. Defaults to 500.
        height (int, optional): Height. Defaults to 500.
        title (str, optional): Title. Defaults to ntk.
        canvas_width (str, optional): Canvas width. Defaults to width.
        canvas_height (str, optional): Canvas height. Defaults to height.
        closing_command (function, optional): Command to execute on close. Defaults to sys.exit.
        resizable (iterable or boolean, optional): Controls whether the window is resizable on the X axis, then Y axis

    Returns:
        _type_: _description_
    """
    # Create window

    if type(resizable) is bool:
        resizable = (resizable, resizable)

    if title is None:
        title = "ntk"

    canvas = _window_internal(
        width,
        height,
        title,
        canvas_width,
        canvas_height,
        closing_command,
        resizable,
        override,
    )

    # Start window thread
    canvas.start()

    # Wait for window to be created, as it == in a separate thread and not blocking this thread
    while canvas.root is None:
        sleep(0.1)

    # Return the window
    return canvas


fonts = [
    "Algerian",
    "Blackadder ITC",
    "Bradley Hand ITC",
    "Castellar",
    "Cooper Black",
    "Edwardian Script ITC",
    "Forte",
    "Comic Sans MS",
    "Bauhaus 93",
    "Harlow Solid Italic",
]

colors = [
    "FD3A4A",
    "FD3A4A",
    "A7F432",
    "5DADEC",
    "BFAFB2",
    "FF5470",
    "FFDB00",
    "FF7A00",
    "#45ba52",
    "#2a89d5",
    "#5c21de",
    "#d14d2e",
]


# NOTE: EXAMPLE WINDOW
def __main__():
    canvas = Window(title=None, width=800, height=500).place(400, 300)
    print(
        fonts_manager.loadfont(
            r"C:\Users\reube\nebulatk\examples\Fonts\HARLOWSI.TTF", private=False
        )
    )
    Frame(canvas, image="examples/Images/background.png", width=500, height=500).place(
        0, 0
    )
    # Button(canvas,10,10,text="hahah").place()
    # Button(canvas,text="hahah").place(50,10)
    # Button(canvas,text="hihih", font = ("Helvetica",36)).place(100,100)
    img = image_manager.Image("examples/Images/main_button_inactive.png")
    btn = Button(
        canvas,
        image=img,
        active_image="examples/Images/main_button_inactive2.png",
        hover_image="examples/Images/main_button_active.png",
        active_hover_image="examples/Images/main_button_active2.png",
        width=100,
        height=100,
        mode="toggle",
        border_width=2,
    )
    btn.place(0, 0)
    btn.place(100, 200)
    btn = Button(
        canvas,
        image=img,
        active_image="examples/Images/main_button_inactive2.png",
        hover_image="examples/Images/main_button_active.png",
        active_hover_image="examples/Images/main_button_active2.png",
        width=300,
        height=300,
        mode="toggle",
        border_width=2,
    )
    btn.place(0, 0)
    btn.place(200, 200)
    img.flip().flip("vertical")
    btn = Button(
        canvas,
        image=img,
        active_image="examples/Images/main_button_inactive2.png",
        hover_image="examples/Images/main_button_active.png",
        active_hover_image="examples/Images/main_button_active2.png",
        width=100,
        height=100,
        mode="toggle",
        border_width=2,
    )
    btn.place(0, 0)
    btn.place(300, 200)
    btn = (
        Button(
            canvas,
            text="hi",
            font=("Harlow Solid Italic", 50),
            fill=[255, 67, 67, 45],
            border_width=2,
        )
        .place(0, 400)
        .place(100, 400)
    )
    btn.text = "hello"
    btn.update()
    Slider(
        canvas,
        width=100,
        height=20,
        slider_width=20,
        slider_height=20,
        slider_border_width=2,
        slider_fill="#ffaaaa",
    ).place(100, 100)

    Entry(
        canvas,
        text="hi",
        font=("Helvetica", 50),
        fill=[255, 67, 67, 45],
        border_width=2,
    ).place(0, 400).place(200, 400)
    # ImageButton(canvas,image="Images/test_inactive.png",active_image="Images/test_active.png").place(0,0)
    # btn_4 = Button(canvas,15,15,text="hahah").place(15,15)
    # btn_4.place(50,60)

    # Frame(canvas,30,30, border = "green").place(160,80)
    print(canvas.children)
    print(standard_methods.get_rect_points(btn))
    # canvas.destroy()
    window = Window(width=800, height=600)

    def animate_btn():
        btn6.width = 100
        btn6.height = 100
        btn6.place(0, 0)
        btn6.update()
        keyframes = [
            (
                1.0,
                {"x": 50, "y": 50},
                animation_controller.Curves.ease_in_quad,
            ),  # Move to (50, 50) in 1s
        ]
        anim_group = animation_controller.AnimationGroup(
            btn6, keyframes, steps=60, looping=True
        )
        anim_group.start()

    btn6 = Button(window, image=img)
    btn6.place(0, 0)
    sleep(2)
    window.resize(1000, 800)  # Change size
    window.configure(title="New Title", resizable=(False, False))  # Change properties
    window.hide()  # Hide window
    sleep(4)
    window.show()
    window.taskbar_manager._initialize_custom_thumbnails()
    animate_btn()
    btn7 = Button(
        canvas,
        text="Test",
        width=100,
        height=100,
        fill="#FF0000",
        hover_fill="#00FF00",
        active_hover_fill="#0000FF",
        active_fill="#000000",
        mode="toggle",
    )
    btn7.place(0, 0)
    # window = Window()

    """def trigger_custom_thumbnail():
        window.taskbar_manager.SetThumbnailClip(0, 0, 100, 100)

    btn = Button(
        window,
        width=100,
        height=100,
        mode="toggle",
        border_width=2,
        command=trigger_custom_thumbnail,
    )
    btn.place(0, 0)

    def animate_btn():
        btn.width = 100
        btn.height = 100
        btn.place(0, 0)
        btn.update()
        keyframes = [
            (
                1.0,
                {"x": 50, "y": 50},
                animation_controller.Curves.ease_in_quad,
            ),  # Move to (50, 50) in 1s
            (
                1,
                {"x": 100, "y": 100},
                animation_controller.Curves.bounce,
                1,
            ),  # Then to (100, 100) in 0.5s
            animation_controller.Animation(
                btn,
                {"x": 0, "y": 0},
                1,
                animation_controller.Curves.ease_out_cubic,
            ),  # Back to (0, 0) in 1s
            [animation_controller.Animation(btn, {"width": 50}, 2.5), 0],
        ]
        anim_group = animation_controller.AnimationGroup(btn, keyframes, steps=60)
        anim_group.start()

    btn2 = Button(
        window,
        text="animate",
        width=100,
        height=50,
        command=animate_btn,
    )
    btn2.place(100, 0)

    anim2 = animation_controller.AnimationGroup(
        btn2,
        [
            [animation_controller.Animation(btn2, {"width": 50}, 1), 0],
            [animation_controller.Animation(btn2, {"height": 100}, 1), 0],
        ],
        steps=60,
        looping=True,
    )
    anim2.start()
    Entry(window, width=100, height=50, text="hello").place(0, 100)

    widget = Button(
        window,
        text="hello",
        width=100,
        height=100,
        fill="#FF0000",
    ).place(
        200, 200
    )  # Red
    anim = animation_controller.Animation(
        widget, {"fill": "#00FF0000"}, duration=1.0, looping=True
    )  # Animate to green
    anim.start()

    container = Container(window, width=200, height=50, fill="#FF0000")
    container.place(20, 10)
    print(container.maps)

    button5 = Button(container, text="hello", width=100, height=100, fill="#FF0000")
    button5.place(0, 0)

    window.taskbar_manager.SetProgress(20)
    window.taskbar_manager.SetThumbnailNotification("info")
    # Add buttons to taskbar
    window.taskbar_manager.AddMediaControlButtons({})

    # Update play/pause state
    window.taskbar_manager.UpdatePlayPauseButton(True)  # Show pause icon

    # Disable next button if at end of playlist
    window.taskbar_manager.SetButtonEnabled(
        taskbar_manager.WindowsConstants.THUMB_BUTTON_FORWARD, False
    )"""


if __name__ == "__main__":
    __main__()



import sys
import os
from unittest.mock import patch
import pytest
import importlib
import types

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../nebulatk"))
)

# Import the fonts_manager module
import fonts_manager


def test_create_font():
    assert fonts_manager.Font(None).font == ("Helvetica", -1)

    assert fonts_manager.Font("Arial").font == ("Arial", 10)


@pytest.mark.skipif(sys.platform != "win32", reason="windll only on Windows")
@patch("ctypes.windll.gdi32.AddFontResourceExW")
def test_load_font(mock_windll):
    assert fonts_manager.ctypes_available
    assert fonts_manager.loadfont("test.ttf")


class FuncStub:
    def __init__(self, fn):
        self._fn = fn

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


class DummyFC:
    def __init__(self):
        self.inited = False
        self.config = object()
        self.add_calls = []
        self.build_calls = []
        self.FcInit = FuncStub(self._fcinit)
        self.FcConfigGetCurrent = FuncStub(self._getcurrent)
        self.FcConfigAppFontAddFile = FuncStub(self._addfile)
        self.FcConfigBuildFonts = FuncStub(self._buildfonts)

    def _fcinit(self):
        self.inited = True
        return True

    def _getcurrent(self):
        return self.config

    def _addfile(self, cfg, path_bytes):
        self.add_calls.append((cfg, path_bytes))
        return 1  # non-zero == success

    def _buildfonts(self, cfg):
        self.build_calls.append(cfg)
        return True


def test_loadfont_linux_branch(monkeypatch):
    # A container to hold the DummyFC instance when LoadLibrary is called:
    created = {}

    # Build a stub 'ctypes' module
    stub_ctypes = types.ModuleType("ctypes")

    # Replace LoadLibrary with one that saves its DummyFC and returns it
    def fake_load(name):
        inst = DummyFC()
        created["fc"] = inst
        return inst

    stub_ctypes.cdll = types.SimpleNamespace(LoadLibrary=fake_load)
    stub_ctypes.c_char_p = __import__("ctypes").c_char_p
    stub_ctypes.c_void_p = __import__("ctypes").c_void_p

    # Stub out ctypes.util.find_library
    stub_ctypes_util = types.ModuleType("ctypes.util")
    stub_ctypes_util.find_library = lambda name: "libfontconfig.so.1"

    # Inject our stubs
    monkeypatch.setitem(sys.modules, "ctypes", stub_ctypes)
    monkeypatch.setitem(sys.modules, "ctypes.util", stub_ctypes_util)
    sys.modules.pop("fonts_manager", None)

    # Re-import under test
    fm = importlib.import_module("fonts_manager")
    assert fm.ctypes_available is False

    # Call loadfont — our fake_load will have run, populating `created['fc']`
    ok = fm.loadfont("/tmp/foo.ttf")
    assert ok is True

    # Grab the exact instance that the module used
    dummy = created["fc"]
    # Now assert directly on *our* dummy
    assert dummy.add_calls == [(dummy.config, b"/tmp/foo.ttf")]
    assert dummy.build_calls == [dummy.config]



import sys
import os
import pytest
from unittest.mock import MagicMock, patch
import types

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../nebulatk"))
)


class MockWindow:
    """Mock window object for testing."""

    def __init__(self):
        self.root = None
        self._setup_mock_root()
        self.width = 1200
        self.height = 900

    def _setup_mock_root(self):
        """Set up mock tkinter root."""
        self.root = MagicMock()
        self.root.winfo_id.return_value = 12345
        self.root.after.return_value = "timer_id"
        self.root.after_cancel = MagicMock()


class MockCOMTypes:
    """Mock COM types for testing."""

    def __init__(self):
        self.GUID = MagicMock()
        self.CoInitialize = MagicMock()
        self.CoUninitialize = MagicMock()
        self.CoCreateInstance = MagicMock()
        self.HRESULT = MagicMock()
        self.POINTER = MagicMock()
        self.IUnknown = MagicMock()
        self.COMMETHOD = MagicMock()

        # Set up comtypes.client module mock
        self.client = MagicMock()
        self.client.CreateObject = MagicMock()

        # Configure CreateObject to return a comprehensive mock
        mock_com_object = MagicMock()
        mock_com_object.QueryInterface = MagicMock()
        mock_com_object.AddRef = MagicMock(return_value=1)
        mock_com_object.Release = MagicMock(return_value=0)

        # Mock common COM interfaces that might be requested
        mock_com_object.ITaskbarList3 = MagicMock()
        mock_com_object.ITaskbarList4 = MagicMock()

        self.client.CreateObject.return_value = mock_com_object


class MockCTypes:
    """Mock ctypes module for testing."""

    def __init__(self):
        self.windll = MagicMock()
        self.wintypes = MagicMock()
        self.CFUNCTYPE = MagicMock()
        self.c_int = int
        self.c_void_p = MagicMock
        self.c_long = int
        self.c_ulong = int  # Added missing c_ulong
        self.c_uint = int
        self.c_ulonglong = int
        self.c_int64 = int
        self.c_uint32 = int
        self.c_int32 = int
        self.c_uint16 = int
        self.c_char_p = MagicMock

        # Create a mock class that supports array syntax (c_type * size)
        class MockCType:
            def __mul__(self, other):
                return MagicMock()  # Return a mock array type

        self.c_wchar = MockCType()  # Fixed c_wchar to support array syntax
        self.c_bool = bool  # Added missing c_bool
        self.byref = MagicMock()
        # This will be set up later with the interface mock
        self.cast = None
        self.POINTER = MagicMock()
        self.Structure = MagicMock()
        self.WINFUNCTYPE = MagicMock()
        self.HRESULT = MagicMock()
        self.sizeof = MagicMock(return_value=8)
        self.create_string_buffer = MagicMock()  # Added missing create_string_buffer

        # Set up OleDLL mock for COM interface creation
        mock_ole32 = MagicMock()
        mock_ole32.CoCreateInstance.return_value = 0  # S_OK
        self.OleDLL = MagicMock(return_value=mock_ole32)

        # Mock vtbl and interface for successful initialization
        mock_vtbl = MagicMock()
        mock_vtbl.HrInit.return_value = 0  # S_OK
        mock_vtbl.SetProgressState.return_value = 0
        mock_vtbl.SetProgressValue.return_value = 0
        mock_vtbl.SetOverlayIcon.return_value = 0
        mock_vtbl.SetThumbnailToolTip.return_value = 0
        mock_vtbl.SetThumbnailClip.return_value = 0

        mock_interface_contents = MagicMock()
        mock_interface_contents.lpVtbl.contents = mock_vtbl

        mock_interface = MagicMock()
        mock_interface.contents = mock_interface_contents

        self.mock_taskbar_interface = mock_interface
        self.mock_vtbl = mock_vtbl

        # Set up cast to return our mock interface
        self.cast = MagicMock(return_value=mock_interface)


@pytest.fixture
def mock_modules():
    """Set up comprehensive mocking for Windows-specific modules."""
    mock_ctypes = MockCTypes()
    mock_ctypes.wintypes.HWND = int
    mock_ctypes.wintypes.POINT = MagicMock
    mock_ctypes.wintypes.HICON = int  # Added missing HICON
    mock_ctypes.wintypes.HBITMAP = int  # Added missing HBITMAP
    mock_ctypes.wintypes.HDC = int  # Added missing HDC
    mock_ctypes.wintypes.HGDIOBJ = int  # Added missing HGDIOBJ
    mock_ctypes.wintypes.BOOL = bool  # Added missing BOOL
    mock_ctypes.wintypes.DWORD = int  # Added missing DWORD
    mock_ctypes.wintypes.UINT = int  # Added missing UINT
    mock_ctypes.wintypes.WPARAM = int  # Added missing WPARAM
    mock_ctypes.wintypes.LPARAM = int  # Added missing LPARAM
    mock_ctypes.wintypes.LPCWSTR = MagicMock  # Added missing LPCWSTR
    mock_ctypes.wintypes.LPWSTR = MagicMock  # Added missing LPWSTR
    mock_ctypes.wintypes.HANDLE = int  # Added missing HANDLE

    # Create a proper mock RECT class that has attributes
    class MockRECT:
        def __init__(self, left=0, top=0, right=0, bottom=0):
            self.left = left
            self.top = top
            self.right = right
            self.bottom = bottom

    mock_ctypes.wintypes.RECT = MockRECT

    mock_comtypes = MockCOMTypes()

    with patch.dict(
        "sys.modules",
        {
            "comtypes": mock_comtypes,
            "comtypes.client": mock_comtypes.client,
            "ctypes": mock_ctypes,
            "ctypes.wintypes": mock_ctypes.wintypes,
        },
    ):
        yield mock_ctypes


@pytest.fixture
def mock_window():
    """Create a mock window for testing."""
    return MockWindow()


class TestTaskbarManagerInit:
    """Test TaskbarManager initialization."""

    @pytest.mark.skipif(
        sys.platform != "win32", reason="Windows-specific functionality"
    )
    def test_init_on_windows(self, mock_modules, mock_window):
        """Test TaskbarManager initialization on Windows."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        assert manager.window is mock_window
        assert manager.auto_invalidate_enabled is True
        assert manager.debounce_delay == 50

    def test_init_cross_platform(self, mock_modules, mock_window):
        """Test TaskbarManager initialization with mocked Windows components."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        assert manager.window is mock_window
        assert manager.debounce_delay == 50
        assert manager.auto_invalidate_enabled is True


class TestTaskbarManagerFeatures:
    """Test TaskbarManager feature methods."""

    def test_set_progress_valid_range(self, mock_modules, mock_window):
        """Test setting progress with valid values and verify correct parameters."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.vtbl.SetProgressState.return_value = 0
        manager.vtbl.SetProgressValue.return_value = 0
        manager.hwnd = 12345

        # Test progress = 0 (should set NOPROGRESS state)
        result = manager.SetProgress(0)
        assert result is True
        manager.vtbl.SetProgressState.assert_called_with(
            manager.taskbar_interface, 12345, 0  # TBPF_NOPROGRESS = 0
        )

        # Reset mocks
        manager.vtbl.SetProgressState.reset_mock()
        manager.vtbl.SetProgressValue.reset_mock()

        # Test progress = 50 (should set NORMAL state and value)
        result = manager.SetProgress(50)
        assert result is True
        manager.vtbl.SetProgressState.assert_called_with(
            manager.taskbar_interface, 12345, 2  # TBPF_NORMAL = 2
        )
        manager.vtbl.SetProgressValue.assert_called_with(
            manager.taskbar_interface, 12345, 50, 100
        )

        # Reset mocks
        manager.vtbl.SetProgressState.reset_mock()
        manager.vtbl.SetProgressValue.reset_mock()

        # Test progress = 100 (should set NORMAL state and value)
        result = manager.SetProgress(100)
        assert result is True
        manager.vtbl.SetProgressState.assert_called_with(
            manager.taskbar_interface, 12345, 2  # TBPF_NORMAL = 2
        )
        manager.vtbl.SetProgressValue.assert_called_with(
            manager.taskbar_interface, 12345, 100, 100
        )

    def test_set_progress_invalid_range(self, mock_modules, mock_window):
        """Test setting progress with invalid values."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Test invalid progress values
        with pytest.raises(ValueError, match="Progress must be between 0 and 100"):
            manager.SetProgress(-1)

        with pytest.raises(ValueError, match="Progress must be between 0 and 100"):
            manager.SetProgress(101)

    def test_set_progress_error_handling(self, mock_modules, mock_window):
        """Test progress setting error handling."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.hwnd = 12345

        # Test SetProgressState failure
        manager.vtbl.SetProgressState.return_value = 0x80004005  # E_FAIL
        result = manager.SetProgress(50)
        assert result is False

        # Test SetProgressValue failure
        manager.vtbl.SetProgressState.return_value = 0  # S_OK
        manager.vtbl.SetProgressValue.return_value = 0x80004005  # E_FAIL
        result = manager.SetProgress(50)
        assert result is False

    def test_set_thumbnail_notification_parameters(self, mock_modules, mock_window):
        """Test thumbnail notification with detailed parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.vtbl.SetOverlayIcon.return_value = 0
        manager.hwnd = 12345

        # Mock LoadIconW to return specific handles
        with patch("ctypes.windll.user32.LoadIconW") as mock_load_icon:
            mock_load_icon.return_value = 9999

            # Test warning icon
            result = manager.SetThumbnailNotification("warning")
            assert result is True

            # Verify LoadIconW was called with correct parameters
            mock_load_icon.assert_called_once()
            # Verify SetOverlayIcon was called with correct parameters
            manager.vtbl.SetOverlayIcon.assert_called_with(
                manager.taskbar_interface, 12345, 9999, "Warning Icon"
            )

            # Reset mocks
            mock_load_icon.reset_mock()
            manager.vtbl.SetOverlayIcon.reset_mock()

            # Test error icon
            result = manager.SetThumbnailNotification("error")
            assert result is True
            manager.vtbl.SetOverlayIcon.assert_called_with(
                manager.taskbar_interface, 12345, 9999, "Error Icon"
            )

            # Reset mocks
            mock_load_icon.reset_mock()
            manager.vtbl.SetOverlayIcon.reset_mock()

            # Test info icon
            result = manager.SetThumbnailNotification("info")
            assert result is True
            manager.vtbl.SetOverlayIcon.assert_called_with(
                manager.taskbar_interface, 12345, 9999, "Info Icon"
            )

            # Reset mocks
            mock_load_icon.reset_mock()
            manager.vtbl.SetOverlayIcon.reset_mock()

            # Test clearing icon (None)
            result = manager.SetThumbnailNotification(None)
            assert result is True
            # LoadIconW should not be called for None
            mock_load_icon.assert_not_called()
            manager.vtbl.SetOverlayIcon.assert_called_with(
                manager.taskbar_interface, 12345, None, ""
            )

    def test_set_thumbnail_notification_error_handling(self, mock_modules, mock_window):
        """Test thumbnail notification error handling."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.vtbl.SetOverlayIcon.return_value = 0x80004005  # E_FAIL
        manager.hwnd = 12345

        result = manager.SetThumbnailNotification("warning")
        assert result is False

    def test_set_thumbnail_clip_parameters(self, mock_modules, mock_window):
        """Test thumbnail clipping with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Mock dwmapi and DwmSetWindowAttribute for initialization
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0
            mock_dwmapi.DwmSetWindowAttribute.return_value = 0

            # Test setting clip rectangle with specific values
            manager.SetThumbnailClip(10, 20, 300, 400)

            # Verify clip_rect is set with correct values
            assert manager.clip_rect is not None
            assert manager.clip_rect.left == 10
            assert manager.clip_rect.top == 20
            assert manager.clip_rect.right == 310  # left + width
            assert manager.clip_rect.bottom == 420  # top + height

            # Verify custom thumbnails were initialized
            assert manager.custom_thumbnails_initialized is True

            # Verify DwmInvalidateIconicBitmaps was called twice (once during initialization, once after setting clip)
            assert mock_dwmapi.DwmInvalidateIconicBitmaps.call_count == 2
            mock_dwmapi.DwmInvalidateIconicBitmaps.assert_called_with(12345)

    def test_set_thumbnail_clip_validation(self, mock_modules, mock_window):
        """Test thumbnail clipping with dimension validation."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True

        # Mock dwmapi
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0

            # Test with negative coordinates (should be clamped to 0)
            manager.SetThumbnailClip(-10, -20, 300, 400)
            assert manager.clip_rect.left == 0
            assert manager.clip_rect.top == 0
            assert manager.clip_rect.right == 300
            assert manager.clip_rect.bottom == 400

            # Test with zero dimensions (should be clamped to 1)
            manager.SetThumbnailClip(10, 20, 0, 0)
            assert manager.clip_rect.left == 10
            assert manager.clip_rect.top == 20
            assert manager.clip_rect.right == 11  # 10 + 1
            assert manager.clip_rect.bottom == 21  # 20 + 1

            # Test with oversized dimensions (should be clamped)
            manager.SetThumbnailClip(0, 0, 50000, 50000)
            assert manager.clip_rect.left == 0
            assert manager.clip_rect.top == 0
            assert manager.clip_rect.right == 32767  # MAX_DIMENSION
            assert manager.clip_rect.bottom == 32767  # MAX_DIMENSION

    def test_clear_thumbnail_clip_parameters(self, mock_modules, mock_window):
        """Test clearing thumbnail clipping rectangle with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 54321

        # Set initial clip rect
        from ctypes import wintypes

        manager.clip_rect = wintypes.RECT(10, 20, 310, 420)

        # Mock dwmapi and DwmSetWindowAttribute
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0
            mock_dwmapi.DwmSetWindowAttribute.return_value = 0

            # Test clearing clip rectangle
            manager.ClearThumbnailClip()

            # Verify clip_rect is None
            assert manager.clip_rect is None

            # Verify custom thumbnails were initialized
            assert manager.custom_thumbnails_initialized is True

            # Verify DwmInvalidateIconicBitmaps was called twice (once during initialization, once after clearing)
            assert mock_dwmapi.DwmInvalidateIconicBitmaps.call_count == 2
            mock_dwmapi.DwmInvalidateIconicBitmaps.assert_called_with(54321)

    def test_auto_invalidation_control(self, mock_modules, mock_window):
        """Test auto-invalidation enable/disable."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Test enabling auto-invalidation
        manager.SetAutoInvalidation(True)
        assert manager.IsAutoInvalidationEnabled() is True

        # Test disabling auto-invalidation
        manager.SetAutoInvalidation(False)
        assert manager.IsAutoInvalidationEnabled() is False

    def test_debounce_delay_control(self, mock_modules, mock_window):
        """Test debounce delay configuration."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Test setting debounce delay
        manager.SetDebounceDelay(100)
        assert manager.GetDebounceDelay() == 100

        # Test minimum debounce delay
        manager.SetDebounceDelay(5)
        assert manager.GetDebounceDelay() == 10  # Should be clamped to minimum

        # Test chaining
        result = manager.SetDebounceDelay(200)
        assert result is manager  # Should return self for chaining

    def test_custom_thumbnails_not_initialized_by_default(
        self, mock_modules, mock_window
    ):
        """Test that custom thumbnails are not initialized by default."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Verify custom thumbnails are not initialized
        assert manager.custom_thumbnails_initialized is False

        # Test that using only notifications/progress doesn't initialize custom thumbnails
        manager.SetProgress(50)
        assert manager.custom_thumbnails_initialized is False

        manager.SetThumbnailNotification("warning")
        assert manager.custom_thumbnails_initialized is False

    def test_custom_thumbnails_initialized_on_first_clip_call(
        self, mock_modules, mock_window
    ):
        """Test that custom thumbnails are initialized when SetThumbnailClip is called."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Verify custom thumbnails are not initialized initially
        assert manager.custom_thumbnails_initialized is False

        # Mock dwmapi for initialization
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0
            mock_dwmapi.DwmSetWindowAttribute.return_value = 0

            # Call SetThumbnailClip
            manager.SetThumbnailClip(0, 0, 100, 100)

            # Verify custom thumbnails are now initialized
            assert manager.custom_thumbnails_initialized is True

            # Verify DwmSetWindowAttribute was called for custom thumbnail setup
            assert (
                mock_dwmapi.DwmSetWindowAttribute.call_count == 3
            )  # Three attributes set

    def test_clear_thumbnail_clip_keeps_custom_thumbnails_initialized(
        self, mock_modules, mock_window
    ):
        """Test that ClearThumbnailClip doesn't turn off custom thumbnails."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Mock dwmapi for initialization
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0
            mock_dwmapi.DwmSetWindowAttribute.return_value = 0

            # First call to SetThumbnailClip initializes custom thumbnails
            manager.SetThumbnailClip(0, 0, 100, 100)
            assert manager.custom_thumbnails_initialized is True

            # Clear the clip
            manager.ClearThumbnailClip()

            # Verify custom thumbnails remain initialized
            assert manager.custom_thumbnails_initialized is True
            assert manager.clip_rect is None

    def test_custom_thumbnails_initialized_only_once(self, mock_modules, mock_window):
        """Test that custom thumbnails are only initialized once, even with multiple SetThumbnailClip calls."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Mock dwmapi for initialization
        with patch("ctypes.windll.dwmapi") as mock_dwmapi:
            mock_dwmapi.DwmInvalidateIconicBitmaps.return_value = 0
            mock_dwmapi.DwmSetWindowAttribute.return_value = 0

            # First call to SetThumbnailClip initializes custom thumbnails
            manager.SetThumbnailClip(0, 0, 100, 100)
            assert manager.custom_thumbnails_initialized is True

            # Check initialization call count
            initial_setup_calls = mock_dwmapi.DwmSetWindowAttribute.call_count
            initial_invalidate_calls = mock_dwmapi.DwmInvalidateIconicBitmaps.call_count

            # Second call to SetThumbnailClip should not reinitialize
            manager.SetThumbnailClip(10, 10, 200, 200)
            assert manager.custom_thumbnails_initialized is True

            # Setup calls should not have increased (no re-initialization)
            assert mock_dwmapi.DwmSetWindowAttribute.call_count == initial_setup_calls
            # Invalidate calls should have increased by 1 (for the new clip)
            assert (
                mock_dwmapi.DwmInvalidateIconicBitmaps.call_count
                == initial_invalidate_calls + 1
            )

            # Verify new clip rectangle is set
            assert manager.clip_rect.left == 10
            assert manager.clip_rect.top == 10
            assert manager.clip_rect.right == 210  # 10 + 200
            assert manager.clip_rect.bottom == 210  # 10 + 200


class TestTaskbarManagerValidation:
    """Test TaskbarManager validation methods."""

    def test_validate_dimensions(self, mock_modules, mock_window):
        """Test dimension validation and clamping."""
        from taskbar_manager import ValidationHelper

        # Test normal dimensions
        width, height = ValidationHelper.validate_dimensions(800, 600)
        assert width == 800
        assert height == 600

        # Test zero/negative dimensions
        width, height = ValidationHelper.validate_dimensions(0, -100)
        assert width == 1
        assert height == 1

        # Test oversized dimensions
        width, height = ValidationHelper.validate_dimensions(50000, 50000)
        assert width == 32767  # MAX_DIMENSION
        assert height == 32767  # MAX_DIMENSION

    def test_validate_coordinates(self, mock_modules, mock_window):
        """Test coordinate validation and clamping."""
        from taskbar_manager import ValidationHelper

        # Test normal coordinates
        x, y = ValidationHelper.validate_coordinates(100, 200)
        assert x == 100
        assert y == 200

        # Test negative coordinates
        x, y = ValidationHelper.validate_coordinates(-50, -100)
        assert x == 0
        assert y == 0

        # Test with max bounds
        x, y = ValidationHelper.validate_coordinates(1000, 2000, max_x=500, max_y=800)
        assert x == 500
        assert y == 800


class TestTaskbarManagerThumbnails:
    """Test TaskbarManager thumbnail functionality."""

    def test_thumbnail_request_delegation(self, mock_modules, mock_window):
        """Test thumbnail request delegation."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Test thumbnail request delegation
        manager._delegate_thumbnail_request(12345, 200, 150)

        # Verify flags are set
        assert manager.thumbnail_request_flag is True
        assert manager.thumbnail_width == 200
        assert manager.thumbnail_height == 150

    def test_live_preview_request_delegation(self, mock_modules, mock_window):
        """Test live preview request delegation."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Test live preview request delegation
        manager._delegate_live_preview_request(12345)

        # Verify flag is set
        assert manager.live_preview_request_flag is True

    def test_thumbnail_requests_ignored_when_not_initialized(
        self, mock_modules, mock_window
    ):
        """Test that thumbnail requests are ignored when custom thumbnails aren't initialized."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345
        manager.custom_thumbnails_initialized = False  # Not initialized

        # Mock the window message handler and DWM constants
        LRESULT, LONG_PTR, WNDPROCTYPE = 0, MagicMock(), MagicMock()

        with patch("ctypes.windll.user32.CallWindowProcW") as mock_call_proc:
            mock_call_proc.return_value = 0

            # Test thumbnail request message when not initialized
            result = manager._handle_window_message(
                12345,
                0x0323,
                0,
                (150 << 16) | 200,
                999,
                LONG_PTR,  # WM_DWMSENDICONICTHUMBNAIL
            )

            # Should fall through to default handler, not return 0
            assert result == 0  # CallWindowProcW return value
            mock_call_proc.assert_called_once()

            # Reset mock
            mock_call_proc.reset_mock()

            # Test live preview request message when not initialized
            result = manager._handle_window_message(
                12345, 0x0326, 0, 0, 999, LONG_PTR  # WM_DWMSENDICONICLIVEPREVIEWBITMAP
            )

            # Should fall through to default handler, not return 0
            assert result == 0  # CallWindowProcW return value
            mock_call_proc.assert_called_once()

    def test_thumbnail_requests_handled_when_initialized(
        self, mock_modules, mock_window
    ):
        """Test that thumbnail requests are handled when custom thumbnails are initialized."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345
        manager.custom_thumbnails_initialized = True  # Initialized

        # Mock the window message handler and DWM constants
        LRESULT, LONG_PTR, WNDPROCTYPE = 0, MagicMock(), MagicMock()

        with patch("ctypes.windll.user32.CallWindowProcW") as mock_call_proc:
            mock_call_proc.return_value = 0

            # Test thumbnail request message when initialized
            result = manager._handle_window_message(
                12345,
                0x0323,
                0,
                (150 << 16) | 200,
                999,
                LONG_PTR,  # WM_DWMSENDICONICTHUMBNAIL
            )

            # Should handle the message and return 0 immediately
            assert result == 0
            mock_call_proc.assert_not_called()  # Should not call default handler

            # Verify the request was delegated
            assert manager.thumbnail_request_flag is True
            assert manager.thumbnail_width == 150
            assert manager.thumbnail_height == 200

            # Reset flags
            manager.thumbnail_request_flag = False

            # Test live preview request message when initialized
            result = manager._handle_window_message(
                12345, 0x0326, 0, 0, 999, LONG_PTR  # WM_DWMSENDICONICLIVEPREVIEWBITMAP
            )

            # Should handle the message and return 0 immediately
            assert result == 0
            mock_call_proc.assert_not_called()  # Should not call default handler

            # Verify the request was delegated
            assert manager.live_preview_request_flag is True

    def test_create_full_thumbnail(self, mock_modules, mock_window):
        """Test full thumbnail creation."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.bitmap_creator = MagicMock()

        # Mock bitmap creator's create_thumbnail_bitmap method
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = 3001

        # Test full thumbnail creation
        result = manager._create_full_thumbnail(200, 150)

        # Verify bitmap creation was called with correct parameters
        manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
            0, 0, None, None, 200, 150, bypass_clipping=True
        )
        assert result == 3001

    def test_handle_thumbnail_request_safe_parameters(self, mock_modules, mock_window):
        """Test safe thumbnail request handling with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.clip_rect = None
        manager.hwnd = 12345
        manager.bitmap_creator = MagicMock()

        # Mock bitmap creation
        with patch.object(manager, "_create_full_thumbnail") as mock_create:
            mock_create.return_value = 3001  # Mock bitmap handle

            # Mock DwmSetIconicThumbnail
            with patch("ctypes.windll.dwmapi.DwmSetIconicThumbnail") as mock_dwm:
                mock_dwm.return_value = 0

                # Mock DeleteObject
                with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                    mock_delete.return_value = True

                    # Test thumbnail request handling
                    manager._handle_thumbnail_request_safe(12345, 200, 150)

                    # Verify thumbnail creation was called with validated dimensions
                    mock_create.assert_called_once_with(200, 150)

                    # Verify DwmSetIconicThumbnail was called with correct parameters
                    mock_dwm.assert_called_once_with(12345, 3001, 0)

                    # Verify bitmap was cleaned up
                    mock_delete.assert_called_once_with(3001)

    def test_handle_thumbnail_request_safe_with_clipping(
        self, mock_modules, mock_window
    ):
        """Test safe thumbnail request handling with clipping enabled."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345
        manager.bitmap_creator = MagicMock()

        # Set up clipping rectangle
        from ctypes import wintypes

        manager.clip_rect = wintypes.RECT(10, 20, 310, 420)

        # Mock bitmap creation
        with patch.object(manager, "_create_clipped_thumbnail") as mock_create:
            mock_create.return_value = 3002  # Mock bitmap handle

            # Mock DwmSetIconicThumbnail
            with patch("ctypes.windll.dwmapi.DwmSetIconicThumbnail") as mock_dwm:
                mock_dwm.return_value = 0

                # Mock DeleteObject
                with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                    mock_delete.return_value = True

                    # Test thumbnail request handling with clipping
                    manager._handle_thumbnail_request_safe(12345, 200, 150)

                    # Verify clipped thumbnail creation was called
                    mock_create.assert_called_once_with(200, 150)

                    # Verify DwmSetIconicThumbnail was called with correct parameters
                    mock_dwm.assert_called_once_with(12345, 3002, 0)

                    # Verify bitmap was cleaned up
                    mock_delete.assert_called_once_with(3002)

    def test_create_clipped_thumbnail_parameters(self, mock_modules, mock_window):
        """Test clipped thumbnail creation with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.bitmap_creator = MagicMock()

        # Set up clipping rectangle
        from ctypes import wintypes

        manager.clip_rect = wintypes.RECT(10, 20, 310, 420)

        # Mock bitmap creator's create_thumbnail_bitmap method
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = 3004

        # Test clipped thumbnail creation
        result = manager._create_clipped_thumbnail(200, 150)

        # Verify bitmap creation was called with correct clip parameters
        manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
            10,  # clip_rect.left
            20,  # clip_rect.top
            300,  # clip_rect.right - clip_rect.left
            400,  # clip_rect.bottom - clip_rect.top
            200,  # thumb_width
            150,  # thumb_height
            clip_rect=manager.clip_rect,
        )
        assert result == 3004

    def test_create_live_preview_bitmap_parameters(self, mock_modules, mock_window):
        """Test live preview bitmap creation with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345
        manager.bitmap_creator = MagicMock()

        # Mock bitmap creator's create_thumbnail_bitmap method
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = 3005

        # Test live preview bitmap creation
        result = manager._create_live_preview_bitmap(400, 300)

        # Verify bitmap creation was called with bypass_clipping=True
        manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
            0, 0, 400, 300, 400, 300, bypass_clipping=True
        )
        assert result == 3005

    def test_handle_live_preview_request_safe_parameters(
        self, mock_modules, mock_window
    ):
        """Test safe live preview request handling with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345
        manager.bitmap_creator = MagicMock()

        # Mock client rect
        mock_rect = MagicMock()
        mock_rect.right = 800
        mock_rect.left = 0
        mock_rect.bottom = 600
        mock_rect.top = 0

        # Mock bitmap creation
        with patch.object(manager, "_create_live_preview_bitmap") as mock_create:
            mock_create.return_value = 3003  # Mock bitmap handle

            # Mock DwmSetIconicLivePreviewBitmap
            with patch(
                "ctypes.windll.dwmapi.DwmSetIconicLivePreviewBitmap"
            ) as mock_dwm:
                mock_dwm.return_value = 0

                # Mock GetClientRect
                with patch("ctypes.windll.user32.GetClientRect") as mock_get_rect:
                    mock_get_rect.return_value = True

                    # Mock wintypes.RECT instantiation
                    with patch("ctypes.wintypes.RECT", return_value=mock_rect):
                        # Mock DeleteObject
                        with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                            mock_delete.return_value = True

                            # Test live preview request handling
                            manager._handle_live_preview_request_safe(12345)

                            # Verify GetClientRect was called with correct hwnd
                            mock_get_rect.assert_called_once()

                            # Verify bitmap creation was called with client dimensions
                            mock_create.assert_called_once_with(800, 600)

                            # Verify DwmSetIconicLivePreviewBitmap was called with correct parameters
                            mock_dwm.assert_called_once_with(12345, 3003, None, 0)

                            # Verify bitmap was cleaned up
                            mock_delete.assert_called_once_with(3003)


class TestTaskbarManagerTooltips:
    """Test TaskbarManager tooltip functionality."""

    def test_set_thumbnail_tooltip_parameters(self, mock_modules, mock_window):
        """Test setting thumbnail tooltip with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.vtbl.SetThumbnailToolTip.return_value = 0
        manager.hwnd = 12345

        # Test setting tooltip
        result = manager._set_thumbnail_tooltip("Test Tooltip")

        # Verify SetThumbnailToolTip was called with correct parameters
        manager.vtbl.SetThumbnailToolTip.assert_called_once_with(
            manager.taskbar_interface, 12345, "Test Tooltip"
        )
        assert result is True

    def test_set_thumbnail_tooltip_error_handling(self, mock_modules, mock_window):
        """Test tooltip setting error handling."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_interface = MagicMock()
        manager.vtbl = MagicMock()
        manager.vtbl.SetThumbnailToolTip.return_value = 0x80004005  # E_FAIL
        manager.hwnd = 12345

        # Test setting tooltip with error
        result = manager._set_thumbnail_tooltip("Test Tooltip")

        # Verify error handling
        assert result is False


class TestTaskbarManagerWindowMessages:
    """Test TaskbarManager window message handling."""

    def test_taskbar_button_created_handling(self, mock_modules, mock_window):
        """Test taskbar button created message handling."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)

        # Mock _apply_taskbar_features
        with patch.object(manager, "_apply_taskbar_features") as mock_apply:
            # Test taskbar button created handling
            manager._on_taskbar_button_created()

            # Verify features were applied
            mock_apply.assert_called_once()

    def test_apply_taskbar_features_sequence(self, mock_modules, mock_window):
        """Test the sequence of basic taskbar feature application (no custom thumbnails)."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345

        # Mock the individual setup methods (should not be called in basic apply)
        with patch.object(manager, "_setup_custom_thumbnails") as mock_setup:
            with patch.object(manager, "_set_thumbnail_tooltip") as mock_tooltip:
                with patch(
                    "ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps"
                ) as mock_invalidate:
                    mock_setup.return_value = True
                    mock_tooltip.return_value = True
                    mock_invalidate.return_value = 0

                    # Test applying taskbar features
                    manager._apply_taskbar_features()

                    # Verify setup sequence does NOT include custom thumbnails
                    mock_setup.assert_not_called()
                    mock_tooltip.assert_not_called()
                    mock_invalidate.assert_not_called()

                    # Verify taskbar button is marked as created
                    assert manager.taskbar_button_created is True
                    # Verify custom thumbnails are NOT initialized
                    assert manager.custom_thumbnails_initialized is False

    def test_setup_custom_thumbnails_parameters(self, mock_modules, mock_window):
        """Test custom thumbnail setup with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345

        # Mock DwmSetWindowAttribute
        with patch("ctypes.windll.dwmapi.DwmSetWindowAttribute") as mock_dwm:
            mock_dwm.return_value = 0  # S_OK

            # Test custom thumbnail setup
            result = manager._setup_custom_thumbnails()

            # Verify DwmSetWindowAttribute was called with correct parameters
            assert mock_dwm.call_count == 3  # Three different attributes

            # Check the calls were made with correct attribute constants
            calls = mock_dwm.call_args_list

            # DWMWA_FORCE_ICONIC_REPRESENTATION = 7
            assert calls[0][0][1] == 7
            # DWMWA_HAS_ICONIC_BITMAP = 10
            assert calls[1][0][1] == 10
            # DWMWA_DISALLOW_PEEK = 12
            assert calls[2][0][1] == 12

            # All calls should be for the correct window handle
            for call in calls:
                assert call[0][0] == 12345

            assert result is True

    def test_manual_apply_when_not_created(self, mock_modules, mock_window):
        """Test manual apply when taskbar button not yet created."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = False

        # Mock _apply_taskbar_features
        with patch.object(manager, "_apply_taskbar_features") as mock_apply:
            # Test manual apply
            manager.manual_apply()

            # Verify features were applied
            mock_apply.assert_called_once()

    def test_manual_apply_when_already_created(self, mock_modules, mock_window):
        """Test manual apply when taskbar button already created."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True

        # Mock _apply_taskbar_features
        with patch.object(manager, "_apply_taskbar_features") as mock_apply:
            # Test manual apply
            manager.manual_apply()

            # Verify features were NOT applied again
            mock_apply.assert_not_called()

    def test_initialize_custom_thumbnails_functionality(
        self, mock_modules, mock_window
    ):
        """Test the _initialize_custom_thumbnails method directly."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Mock the individual setup methods
        with patch.object(manager, "_setup_custom_thumbnails") as mock_setup:
            with patch.object(manager, "_set_thumbnail_tooltip") as mock_tooltip:
                with patch(
                    "ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps"
                ) as mock_invalidate:
                    mock_setup.return_value = True
                    mock_tooltip.return_value = True
                    mock_invalidate.return_value = 0

                    # Test initializing custom thumbnails
                    manager._initialize_custom_thumbnails()

                    # Verify setup sequence
                    mock_setup.assert_called_once()
                    mock_tooltip.assert_called_once_with("TaskbarManager Demo")
                    mock_invalidate.assert_called_once_with(12345)

                    # Verify custom thumbnails are marked as initialized
                    assert manager.custom_thumbnails_initialized is True

                    # Test calling it again should not reinitialize
                    mock_setup.reset_mock()
                    mock_tooltip.reset_mock()
                    mock_invalidate.reset_mock()

                    manager._initialize_custom_thumbnails()

                    # Verify no additional calls were made
                    mock_setup.assert_not_called()
                    mock_tooltip.assert_not_called()
                    mock_invalidate.assert_not_called()

                    # Should still be initialized
                    assert manager.custom_thumbnails_initialized is True


class TestTaskbarManagerInvalidation:
    """Test TaskbarManager invalidation functionality."""

    def test_invalidate_live_preview_parameters(self, mock_modules, mock_window):
        """Test live preview invalidation with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.custom_thumbnails_initialized = (
            True  # Must be initialized for invalidation to work
        )
        manager.hwnd = 54321

        # Mock DwmInvalidateIconicBitmaps
        with patch("ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps") as mock_dwm:
            mock_dwm.return_value = 0

            # Test invalidation
            manager.InvalidateLivePreview()

            # Verify DWM API was called with correct window handle
            mock_dwm.assert_called_once_with(54321)

    def test_invalidate_live_preview_when_not_created(self, mock_modules, mock_window):
        """Test live preview invalidation when taskbar button not created."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = False

        # Mock DwmInvalidateIconicBitmaps
        with patch("ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps") as mock_dwm:
            # Test invalidation
            manager.InvalidateLivePreview()

            # Verify DWM API was NOT called
            mock_dwm.assert_not_called()

    def test_invalidate_live_preview_when_custom_thumbnails_not_initialized(
        self, mock_modules, mock_window
    ):
        """Test live preview invalidation when custom thumbnails not initialized."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.custom_thumbnails_initialized = False  # Not initialized
        manager.hwnd = 54321

        # Mock DwmInvalidateIconicBitmaps
        with patch("ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps") as mock_dwm:
            # Test invalidation
            manager.InvalidateLivePreview()

            # Verify DWM API was NOT called
            mock_dwm.assert_not_called()

    def test_auto_invalidate_when_custom_thumbnails_not_initialized(
        self, mock_modules, mock_window
    ):
        """Test auto-invalidation when custom thumbnails not initialized."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.custom_thumbnails_initialized = False  # Not initialized
        manager.debounce_delay = 50

        # Mock root.after to capture the callback
        manager.root.after = MagicMock(return_value="timer_id")
        manager.root.after_cancel = MagicMock()

        # Test auto invalidation
        manager._auto_invalidate_thumbnails()

        # Verify timer was NOT scheduled since custom thumbnails aren't initialized
        manager.root.after.assert_not_called()
        manager.root.after_cancel.assert_not_called()

    def test_auto_invalidate_thumbnails_parameters(self, mock_modules, mock_window):
        """Test automatic thumbnail invalidation with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.custom_thumbnails_initialized = (
            True  # Must be initialized for auto-invalidation to work
        )
        manager.debounce_delay = 50

        # Mock root.after to capture the callback
        manager.root.after = MagicMock(return_value="timer_id")
        manager.root.after_cancel = MagicMock()

        # Test auto invalidation
        manager._auto_invalidate_thumbnails()

        # Verify timer was scheduled with correct delay
        manager.root.after.assert_called_once_with(50, manager._perform_invalidation)

        # Test cancellation of previous timer
        manager._invalidate_pending = "old_timer_id"
        manager._auto_invalidate_thumbnails()

        # Verify old timer was cancelled
        manager.root.after_cancel.assert_called_with("old_timer_id")

    def test_perform_invalidation_parameters(self, mock_modules, mock_window):
        """Test actual invalidation performance with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 98765

        # Mock DwmInvalidateIconicBitmaps
        with patch("ctypes.windll.dwmapi.DwmInvalidateIconicBitmaps") as mock_dwm:
            mock_dwm.return_value = 0

            # Set up pending invalidation attribute
            manager._invalidate_pending = "test_timer_id"

            # Test invalidation
            manager._perform_invalidation()

            # Verify DWM API was called with correct window handle
            mock_dwm.assert_called_once_with(98765)

            # Verify pending invalidation attribute was cleared
            assert not hasattr(manager, "_invalidate_pending")


class TestTaskbarManagerPolling:
    """Test TaskbarManager polling functionality with parameter verification."""

    def test_poll_for_requests_thumbnail_debouncing(self, mock_modules, mock_window):
        """Test polling for thumbnail requests with debouncing parameters."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.thumbnail_request_flag = True
        manager.thumbnail_pending_timer = "old_timer_id"
        manager.debounce_delay = 75

        # Mock root.after and after_cancel
        manager.root.after = MagicMock(return_value="new_timer_id")
        manager.root.after_cancel = MagicMock()

        # Test polling
        manager._poll_for_requests()

        # Verify old timer was cancelled
        manager.root.after_cancel.assert_called_with("old_timer_id")

        # Verify new timer was scheduled with correct delay and callback
        calls = manager.root.after.call_args_list
        assert len(calls) >= 1

        # Find the debounced thumbnail call
        thumbnail_call = None
        for call in calls:
            if call[0][0] == 75:  # debounce_delay
                thumbnail_call = call
                break

        assert thumbnail_call is not None
        assert thumbnail_call[0][0] == 75
        assert thumbnail_call[0][1] == manager._process_debounced_thumbnail

        # Verify flag was cleared
        assert manager.thumbnail_request_flag is False

    def test_poll_for_requests_live_preview_debouncing(self, mock_modules, mock_window):
        """Test polling for live preview requests with debouncing parameters."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.live_preview_request_flag = True
        manager.live_preview_pending_timer = "old_preview_timer"
        manager.debounce_delay = 100

        # Mock root.after and after_cancel
        manager.root.after = MagicMock(return_value="new_preview_timer")
        manager.root.after_cancel = MagicMock()

        # Test polling
        manager._poll_for_requests()

        # Verify old timer was cancelled
        manager.root.after_cancel.assert_called_with("old_preview_timer")

        # Verify new timer was scheduled with correct delay and callback
        calls = manager.root.after.call_args_list
        assert len(calls) >= 1

        # Find the debounced live preview call
        preview_call = None
        for call in calls:
            if call[0][0] == 100:  # debounce_delay
                preview_call = call
                break

        assert preview_call is not None
        assert preview_call[0][0] == 100
        assert preview_call[0][1] == manager._process_debounced_live_preview

        # Verify flag was cleared
        assert manager.live_preview_request_flag is False

    def test_process_debounced_thumbnail_parameters(self, mock_modules, mock_window):
        """Test debounced thumbnail processing with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 11111
        manager.thumbnail_width = 320
        manager.thumbnail_height = 240

        # Mock the actual processing method
        with patch.object(manager, "_handle_thumbnail_request_safe") as mock_handle:
            # Test debounced thumbnail processing
            manager._process_debounced_thumbnail()

            # Verify correct parameters were passed
            mock_handle.assert_called_once_with(11111, 320, 240)

            # Verify timer reference was cleared
            assert manager.thumbnail_pending_timer is None

    def test_process_debounced_live_preview_parameters(self, mock_modules, mock_window):
        """Test debounced live preview processing with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 22222

        # Mock the actual processing method
        with patch.object(manager, "_handle_live_preview_request_safe") as mock_handle:
            # Test debounced live preview processing
            manager._process_debounced_live_preview()

            # Verify correct parameters were passed
            mock_handle.assert_called_once_with(22222)

            # Verify timer reference was cleared
            assert manager.live_preview_pending_timer is None


class TestTaskbarManagerValidationEdgeCases:
    """Test TaskbarManager validation edge cases with detailed parameter verification."""

    def test_validate_dimensions_edge_cases(self, mock_modules, mock_window):
        """Test dimension validation edge cases."""
        from taskbar_manager import ValidationHelper

        # Test float inputs (should be converted to int)
        width, height = ValidationHelper.validate_dimensions(800.7, 600.3)
        assert width == 800
        assert height == 600
        assert isinstance(width, int)
        assert isinstance(height, int)

        # Test very large negative numbers
        width, height = ValidationHelper.validate_dimensions(-99999, -88888)
        assert width == 1
        assert height == 1

        # Test exact boundary values
        width, height = ValidationHelper.validate_dimensions(32767, 32767)
        assert width == 32767
        assert height == 32767

        # Test just over boundary
        width, height = ValidationHelper.validate_dimensions(32768, 32768)
        assert width == 32767  # Should be clamped
        assert height == 32767  # Should be clamped

    def test_validate_coordinates_edge_cases(self, mock_modules, mock_window):
        """Test coordinate validation edge cases."""
        from taskbar_manager import ValidationHelper

        # Test float inputs (should be converted to int)
        x, y = ValidationHelper.validate_coordinates(100.9, 200.1)
        assert x == 100
        assert y == 200
        assert isinstance(x, int)
        assert isinstance(y, int)

        # Test very large negative numbers
        x, y = ValidationHelper.validate_coordinates(-50000, -60000)
        assert x == 0
        assert y == 0

        # Test with max bounds as floats
        x, y = ValidationHelper.validate_coordinates(
            1000.5, 2000.7, max_x=500.2, max_y=800.8
        )
        assert x == 500  # Should respect int(max_x)
        assert y == 800  # Should respect int(max_y)

        # Test exact boundary values
        x, y = ValidationHelper.validate_coordinates(32767, 32767)
        assert x == 32767
        assert y == 32767

        # Test just over boundary
        x, y = ValidationHelper.validate_coordinates(32768, 32768)
        assert x == 32767  # Should be clamped
        assert y == 32767  # Should be clamped

    def test_dimension_validation_with_name_parameter(self, mock_modules, mock_window):
        """Test dimension validation with custom name parameter for error messages."""
        from taskbar_manager import ValidationHelper

        # Test with custom name (should not affect validation logic, just messaging)
        width, height = ValidationHelper.validate_dimensions(
            50000, 40000, "custom region"
        )
        assert width == 32767
        assert height == 32767

        # Test normal case with custom name
        width, height = ValidationHelper.validate_dimensions(800, 600, "test area")
        assert width == 800
        assert height == 600


class TestTaskbarManagerErrorHandling:
    """Test TaskbarManager error handling scenarios."""

    def test_bitmap_creation_failure_handling(self, mock_modules, mock_window):
        """Test handling of bitmap creation failures."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.clip_rect = None
        manager.bitmap_creator = MagicMock()

        # Mock bitmap creation to fail
        with patch.object(manager, "_create_full_thumbnail") as mock_create:
            mock_create.return_value = None  # Simulate failure

            # Mock DwmSetIconicThumbnail (should not be called)
            with patch("ctypes.windll.dwmapi.DwmSetIconicThumbnail") as mock_dwm:
                # Test thumbnail request handling with failed bitmap creation
                manager._handle_thumbnail_request_safe(12345, 200, 150)

                # Verify DwmSetIconicThumbnail was NOT called due to failed bitmap
                mock_dwm.assert_not_called()

    def test_dwm_api_error_handling(self, mock_modules, mock_window):
        """Test handling of DWM API errors."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345

        # Mock DwmSetWindowAttribute to fail
        with patch("ctypes.windll.dwmapi.DwmSetWindowAttribute") as mock_dwm:
            mock_dwm.return_value = 0x80004005  # E_FAIL

            # Test custom thumbnail setup with DWM error
            result = manager._setup_custom_thumbnails()

            # Verify error is handled gracefully
            assert result is False  # Should return False on error

    def test_com_interface_creation_error(self, mock_modules, mock_window):
        """Test handling of COM interface creation errors."""
        # Override the mock to simulate COM creation failure
        mock_ctypes = MockCTypes()
        mock_ole32 = MagicMock()
        mock_ole32.CoCreateInstance.return_value = 0x80004005  # E_FAIL
        mock_ctypes.OleDLL = MagicMock(return_value=mock_ole32)

        with patch.dict(
            "sys.modules",
            {
                "comtypes": MockCOMTypes(),
                "ctypes": mock_ctypes,
                "ctypes.wintypes": mock_ctypes.wintypes,
            },
        ):
            # Clear the module cache to force reimport
            if "taskbar_manager" in sys.modules:
                del sys.modules["taskbar_manager"]

            # Test that COM creation error is handled properly
            from taskbar_manager import TaskbarManager

            with pytest.raises(OSError, match="Failed to create taskbar interface"):
                TaskbarManager(mock_window)

    def test_vtbl_initialization_error(self, mock_modules, mock_window):
        """Test handling of vtbl initialization errors."""
        # Set up mock that succeeds COM creation but fails initialization
        mock_ctypes = MockCTypes()
        mock_ole32 = MagicMock()
        mock_ole32.CoCreateInstance.return_value = 0  # S_OK
        mock_ctypes.OleDLL = MagicMock(return_value=mock_ole32)

        # Mock vtbl initialization to fail
        mock_vtbl = MagicMock()
        mock_vtbl.HrInit.return_value = 0x80004005  # E_FAIL

        mock_interface_contents = MagicMock()
        mock_interface_contents.lpVtbl.contents = mock_vtbl

        mock_interface = MagicMock()
        mock_interface.contents = mock_interface_contents

        mock_ctypes.cast = MagicMock(return_value=mock_interface)

        with patch.dict(
            "sys.modules",
            {
                "comtypes": MockCOMTypes(),
                "ctypes": mock_ctypes,
                "ctypes.wintypes": mock_ctypes.wintypes,
            },
        ):
            # Clear the module cache to force reimport
            if "taskbar_manager" in sys.modules:
                del sys.modules["taskbar_manager"]

            # Test that vtbl initialization error is handled properly
            from taskbar_manager import TaskbarManager

            with pytest.raises(OSError, match="Failed to initialize taskbar interface"):
                TaskbarManager(mock_window)


class TestTaskbarManagerIntegration:
    """Integration tests for TaskbarManager."""

    def test_full_initialization_flow(self, mock_modules, mock_window):
        """Test complete TaskbarManager initialization flow."""
        from taskbar_manager import TaskbarManager

        # Create manager and verify it initializes without errors
        manager = TaskbarManager(mock_window)

        # Verify key components are set up
        assert manager.window is mock_window
        assert manager.auto_invalidate_enabled is True

        # Test operations don't crash
        manager.SetProgress(50)
        manager.SetThumbnailClip(0, 0, 100, 100)
        manager.ClearThumbnailClip()
        manager.SetAutoInvalidation(False)
        manager.SetDebounceDelay(100)

        # Verify the manager is still functional
        assert manager.IsAutoInvalidationEnabled() is False
        assert manager.GetDebounceDelay() == 100

    def test_complete_thumbnail_workflow(self, mock_modules, mock_window):
        """Test complete thumbnail workflow with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 99999
        manager.taskbar_button_created = True
        manager.bitmap_creator = MagicMock()

        # Set up comprehensive mocking for the entire workflow
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = (
            4001  # Mock bitmap handle
        )

        with patch("ctypes.windll.dwmapi.DwmSetIconicThumbnail") as mock_set_thumbnail:
            mock_set_thumbnail.return_value = 0

            with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                mock_delete.return_value = True

                # 1. Set clipping rectangle
                manager.SetThumbnailClip(50, 75, 400, 300)

                # 2. Simulate thumbnail request
                manager._delegate_thumbnail_request(99999, 250, 200)

                # 3. Process the request
                manager._handle_thumbnail_request_safe(99999, 250, 200)

                # Verify the complete workflow
                # Clipping should be set correctly
                assert manager.clip_rect.left == 50
                assert manager.clip_rect.top == 75
                assert manager.clip_rect.right == 450  # 50 + 400
                assert manager.clip_rect.bottom == 375  # 75 + 300

                # Request should be delegated correctly
                assert manager.thumbnail_width == 250
                assert manager.thumbnail_height == 200

                # Bitmap should be created with clipped parameters
                manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
                    50,  # clip_rect.left
                    75,  # clip_rect.top
                    400,  # width
                    300,  # height
                    250,  # thumb_width
                    200,  # thumb_height
                    clip_rect=manager.clip_rect,
                )

                # DWM should be called with correct parameters
                mock_set_thumbnail.assert_called_once_with(99999, 4001, 0)

                # Bitmap should be cleaned up
                mock_delete.assert_called_once_with(4001)

    def test_complete_live_preview_workflow(self, mock_modules, mock_window):
        """Test complete live preview workflow with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 88888
        manager.taskbar_button_created = True
        manager.bitmap_creator = MagicMock()

        # Mock client rect for live preview
        mock_rect = MagicMock()
        mock_rect.right = 1200
        mock_rect.left = 0
        mock_rect.bottom = 900
        mock_rect.top = 0

        # Set up comprehensive mocking for live preview workflow
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = (
            5001  # Mock bitmap handle
        )

        with patch(
            "ctypes.windll.dwmapi.DwmSetIconicLivePreviewBitmap"
        ) as mock_set_preview:
            mock_set_preview.return_value = 0

            with patch("ctypes.windll.user32.GetClientRect") as mock_get_rect:
                mock_get_rect.return_value = True

                with patch("ctypes.wintypes.RECT", return_value=mock_rect):
                    with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                        mock_delete.return_value = True

                        # 1. Simulate live preview request
                        manager._delegate_live_preview_request(88888)

                        # 2. Process the request
                        manager._handle_live_preview_request_safe(88888)

                        # Verify the complete workflow
                        # Request should be delegated correctly
                        assert manager.live_preview_request_flag is True

                        # Client rect should be queried (may be called multiple times due to initialization)
                        assert mock_get_rect.call_count >= 1

                        # Bitmap should be created with bypass_clipping=True
                        manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
                            0, 0, 1200, 900, 1200, 900, bypass_clipping=True
                        )

                        # DWM should be called with correct parameters
                        mock_set_preview.assert_called_once_with(88888, 5001, None, 0)

                        # Bitmap should be cleaned up
                        mock_delete.assert_called_once_with(5001)


class TestTaskbarManagerMediaControls:
    """Test TaskbarManager media control button functionality."""

    def test_add_media_control_buttons_default(self, mock_modules, mock_window):
        """Test adding media control buttons with default icons."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        # Mock the toolbar manager creation
        with patch("taskbar_manager.ThumbnailToolbarManager") as mock_toolbar_class:
            mock_toolbar = MagicMock()
            mock_toolbar.create_media_buttons.return_value = True
            mock_toolbar_class.return_value = mock_toolbar

            callbacks = {WindowsConstants.THUMB_BUTTON_PLAY_PAUSE: lambda x, y: None}

            result = manager.AddMediaControlButtons(callbacks)

            assert result is True
            assert manager.toolbar_manager is mock_toolbar
            mock_toolbar.create_media_buttons.assert_called_once_with(callbacks, None)

    def test_add_media_control_buttons_with_custom_icons(
        self, mock_modules, mock_window
    ):
        """Test adding media control buttons with custom icons."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        with patch("taskbar_manager.ThumbnailToolbarManager") as mock_toolbar_class:
            mock_toolbar = MagicMock()
            mock_toolbar.create_media_buttons.return_value = True
            mock_toolbar_class.return_value = mock_toolbar

            callbacks = {WindowsConstants.THUMB_BUTTON_PLAY_PAUSE: lambda x, y: None}
            custom_icons = {
                "play": "icons/play.ico",
                "pause": ("shell32.dll", 16),
                "forward": 12345,  # HICON handle
            }

            result = manager.AddMediaControlButtons(callbacks, custom_icons)

            assert result is True
            mock_toolbar.create_media_buttons.assert_called_once_with(
                callbacks, custom_icons
            )

    def test_add_media_control_buttons_before_taskbar_ready(
        self, mock_modules, mock_window
    ):
        """Test adding media control buttons before taskbar button is created."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = False

        callbacks = {WindowsConstants.THUMB_BUTTON_PLAY_PAUSE: lambda x, y: None}
        custom_icons = {"play": "icons/play.ico"}

        result = manager.AddMediaControlButtons(callbacks, custom_icons)

        assert result is False
        assert manager._pending_button_callbacks is callbacks
        assert manager._pending_custom_icons is custom_icons

    def test_add_media_control_buttons_creation_failure(
        self, mock_modules, mock_window
    ):
        """Test handling of media control button creation failure."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = True
        manager.hwnd = 12345

        with patch("taskbar_manager.ThumbnailToolbarManager") as mock_toolbar_class:
            mock_toolbar = MagicMock()
            mock_toolbar.create_media_buttons.return_value = False
            mock_toolbar_class.return_value = mock_toolbar

            result = manager.AddMediaControlButtons()

            assert result is False

    def test_update_play_pause_button(self, mock_modules, mock_window):
        """Test updating play/pause button state."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        mock_toolbar = MagicMock()
        mock_toolbar.update_play_pause_button.return_value = True
        manager.toolbar_manager = mock_toolbar

        result = manager.UpdatePlayPauseButton(True)

        assert result is True
        mock_toolbar.update_play_pause_button.assert_called_once_with(True)

    def test_update_play_pause_button_not_initialized(self, mock_modules, mock_window):
        """Test updating play/pause button when not initialized."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.toolbar_manager = None

        result = manager.UpdatePlayPauseButton(True)

        assert result is False

    def test_set_button_enabled(self, mock_modules, mock_window):
        """Test enabling/disabling buttons."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        mock_toolbar = MagicMock()
        mock_toolbar.set_button_enabled.return_value = True
        manager.toolbar_manager = mock_toolbar

        result = manager.SetButtonEnabled(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, False
        )

        assert result is True
        mock_toolbar.set_button_enabled.assert_called_once_with(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, False
        )

    def test_set_button_enabled_not_initialized(self, mock_modules, mock_window):
        """Test setting button enabled state when not initialized."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.toolbar_manager = None

        result = manager.SetButtonEnabled(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, False
        )

        assert result is False

    def test_update_button_icon(self, mock_modules, mock_window):
        """Test updating button icon."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        mock_toolbar = MagicMock()
        mock_toolbar.update_button_icon.return_value = True
        manager.toolbar_manager = mock_toolbar

        icon_source = "icons/new_play.ico"
        result = manager.UpdateButtonIcon(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, icon_source
        )

        assert result is True
        mock_toolbar.update_button_icon.assert_called_once_with(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, icon_source
        )

    def test_update_button_icon_not_initialized(self, mock_modules, mock_window):
        """Test updating button icon when not initialized."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.toolbar_manager = None

        result = manager.UpdateButtonIcon(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, "icon.ico"
        )

        assert result is False

    def test_get_available_system_icons(self, mock_modules, mock_window):
        """Test getting available system icons."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        icons = manager.GetAvailableSystemIcons()

        assert isinstance(icons, dict)
        assert len(icons) > 0

        # Check that some expected icons are present
        expected_icons = [
            "media_play",
            "media_pause",
            "media_stop",
            "arrow_left",
            "arrow_right",
        ]
        for icon_name in expected_icons:
            assert icon_name in icons
            assert isinstance(icons[icon_name], tuple)
            assert len(icons[icon_name]) == 2  # (dll_name, icon_index)

    def test_get_button_constants(self, mock_modules, mock_window):
        """Test getting button constants."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        constants = manager.GetButtonConstants()

        assert isinstance(constants, dict)
        assert constants["BACKWARD"] == WindowsConstants.THUMB_BUTTON_BACKWARD
        assert constants["PLAY_PAUSE"] == WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        assert constants["FORWARD"] == WindowsConstants.THUMB_BUTTON_FORWARD
        assert constants["STOP"] == WindowsConstants.THUMB_BUTTON_STOP

    def test_pending_buttons_applied_on_taskbar_creation(
        self, mock_modules, mock_window
    ):
        """Test that pending button callbacks are applied when taskbar is created."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.taskbar_button_created = False

        # Set up pending callbacks and icons
        callbacks = {WindowsConstants.THUMB_BUTTON_PLAY_PAUSE: lambda x, y: None}
        custom_icons = {"play": "icon.ico"}

        manager.AddMediaControlButtons(callbacks, custom_icons)
        assert manager._pending_button_callbacks is callbacks
        assert manager._pending_custom_icons is custom_icons

        # Mock toolbar manager
        with patch("taskbar_manager.ThumbnailToolbarManager") as mock_toolbar_class:
            mock_toolbar = MagicMock()
            mock_toolbar.create_media_buttons.return_value = True
            mock_toolbar_class.return_value = mock_toolbar

            # Simulate taskbar button creation
            manager._apply_taskbar_features()

            assert manager._pending_button_callbacks is None
            assert manager._pending_custom_icons is None
            mock_toolbar.create_media_buttons.assert_called_once_with(
                callbacks, custom_icons
            )


class TestThumbnailToolbarManager:
    """Test ThumbnailToolbarManager functionality."""

    def test_thumbnail_toolbar_manager_init(self, mock_modules, mock_window):
        """Test ThumbnailToolbarManager initialization."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        assert toolbar_manager.taskbar_manager is taskbar_manager
        assert toolbar_manager.buttons == []
        assert toolbar_manager.button_callbacks == {}
        assert toolbar_manager.is_playing is False
        assert toolbar_manager.button_icons == {}

    def test_load_button_icons_default(self, mock_modules, mock_window):
        """Test loading default system icons."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        with patch.object(toolbar_manager, "_load_system_icon") as mock_load_system:
            mock_load_system.return_value = 12345  # Mock HICON

            icons = toolbar_manager._load_button_icons({})

            # Should have loaded all default icons
            assert len(icons) == 5  # backward, play, pause, forward, stop
            assert all(icons[name] == 12345 for name in icons)
            assert mock_load_system.call_count == 5

    def test_load_button_icons_custom(self, mock_modules, mock_window):
        """Test loading custom icons with fallback to defaults."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        custom_icons = {"play": "custom_play.ico", "pause": ("shell32.dll", 16)}

        with (
            patch.object(toolbar_manager, "_load_custom_icon") as mock_load_custom,
            patch.object(toolbar_manager, "_load_system_icon") as mock_load_system,
        ):

            # Custom icon loading succeeds for play, fails for pause
            mock_load_custom.side_effect = [54321, None]  # Success, then failure
            mock_load_system.return_value = 12345  # Default system icon

            icons = toolbar_manager._load_button_icons(custom_icons)

            # Play should use custom icon, others should use system default
            assert icons["play"] == 54321
            assert icons["pause"] == 12345  # Fallback to system
            assert icons["backward"] == 12345  # System default
            assert icons["forward"] == 12345  # System default
            assert icons["stop"] == 12345  # System default

    def test_load_custom_icon_file_path(self, mock_modules, mock_window):
        """Test loading icon from file path."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        with patch.object(toolbar_manager, "_load_icon_from_file") as mock_load_file:
            mock_load_file.return_value = 98765

            result = toolbar_manager._load_custom_icon("icon.ico", "test")

            assert result == 98765
            mock_load_file.assert_called_once_with("icon.ico")

    def test_load_custom_icon_system_dll(self, mock_modules, mock_window):
        """Test loading icon from system DLL."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        with patch.object(toolbar_manager, "_load_system_icon") as mock_load_system:
            mock_load_system.return_value = 11111

            result = toolbar_manager._load_custom_icon(("shell32.dll", 16), "test")

            assert result == 11111
            mock_load_system.assert_called_once_with("shell32.dll", 16)

    def test_load_custom_icon_hicon_handle(self, mock_modules, mock_window):
        """Test loading icon from HICON handle."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        result = toolbar_manager._load_custom_icon(22222, "test")

        assert result == 22222

    def test_load_custom_icon_invalid_source(self, mock_modules, mock_window):
        """Test loading icon with invalid source type."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        result = toolbar_manager._load_custom_icon(["invalid", "type"], "test")

        assert result is None

    @patch("os.path.exists", return_value=True)
    def test_load_icon_from_file_ico(self, mock_exists, mock_modules, mock_window):
        """Test loading .ico file."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.LoadImageW.return_value = 33333
        mock_ctypes.windll.user32 = mock_user32

        result = toolbar_manager._load_icon_from_file("test.ico")

        assert result == 33333
        mock_user32.LoadImageW.assert_called_once()

    @patch("os.path.exists", return_value=True)
    def test_load_icon_from_file_png(self, mock_exists, mock_modules, mock_window):
        """Test loading .png file (converted to icon)."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.LoadImageW.return_value = 44444  # Bitmap handle
        mock_ctypes.windll.user32 = mock_user32

        mock_gdi32 = MagicMock()
        mock_ctypes.windll.gdi32 = mock_gdi32

        with patch.object(toolbar_manager, "_bitmap_to_icon") as mock_bitmap_to_icon:
            mock_bitmap_to_icon.return_value = 55555

            result = toolbar_manager._load_icon_from_file("test.png")

            assert result == 55555
            mock_bitmap_to_icon.assert_called_once_with(44444)
            mock_gdi32.DeleteObject.assert_called_once_with(44444)

    @patch("os.path.exists", return_value=False)
    def test_load_icon_from_file_not_found(
        self, mock_exists, mock_modules, mock_window
    ):
        """Test loading icon from non-existent file."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        result = toolbar_manager._load_icon_from_file("nonexistent.ico")

        assert result is None

    def test_update_play_pause_button_to_playing(self, mock_modules, mock_window):
        """Test updating play/pause button to playing state."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        mock_button.hIcon = 11111
        toolbar_manager.buttons = [mock_button]
        toolbar_manager._icons = {"play": 22222, "pause": 33333}

        result = toolbar_manager.update_play_pause_button(True)

        assert result is True
        assert toolbar_manager.is_playing is True
        assert mock_button.hIcon == 33333  # Pause icon
        assert mock_button.szTip == "Pause"

    def test_update_play_pause_button_to_paused(self, mock_modules, mock_window):
        """Test updating play/pause button to paused state."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)
        toolbar_manager.is_playing = True  # Start in playing state

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        mock_button.hIcon = 33333
        toolbar_manager.buttons = [mock_button]
        toolbar_manager._icons = {"play": 22222, "pause": 33333}

        result = toolbar_manager.update_play_pause_button(False)

        assert result is True
        assert toolbar_manager.is_playing is False
        assert mock_button.hIcon == 22222  # Play icon
        assert mock_button.szTip == "Play"

    def test_set_button_enabled_success(self, mock_modules, mock_window):
        """Test enabling/disabling button successfully."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_FORWARD
        toolbar_manager.buttons = [mock_button]

        result = toolbar_manager.set_button_enabled(
            WindowsConstants.THUMB_BUTTON_FORWARD, False
        )

        assert result is True
        assert mock_button.dwFlags == WindowsConstants.THBF_DISABLED

    def test_set_button_enabled_not_found(self, mock_modules, mock_window):
        """Test enabling/disabling non-existent button."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)
        toolbar_manager.buttons = []

        result = toolbar_manager.set_button_enabled(
            WindowsConstants.THUMB_BUTTON_FORWARD, False
        )

        assert result is False

    def test_update_button_icon_success(self, mock_modules, mock_window):
        """Test updating button icon successfully."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        mock_button.hIcon = 11111  # Old icon
        toolbar_manager.buttons = [mock_button]

        with patch.object(toolbar_manager, "_load_custom_icon") as mock_load:
            mock_load.return_value = 99999  # New icon

            result = toolbar_manager.update_button_icon(
                WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, "new_icon.ico"
            )

            assert result is True
            assert mock_button.hIcon == 99999

    def test_update_button_icon_load_failure(self, mock_modules, mock_window):
        """Test updating button icon when loading fails."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        with patch.object(toolbar_manager, "_load_custom_icon") as mock_load:
            mock_load.return_value = None  # Loading failed

            result = toolbar_manager.update_button_icon(
                WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, "bad_icon.ico"
            )

            assert result is False

    def test_cleanup_resources(self, mock_modules, mock_window):
        """Test cleanup of custom icon resources."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up some custom icons
        toolbar_manager.button_icons["custom_icons"] = [11111, 22222, 33333]

        mock_user32 = MagicMock()
        mock_modules.windll.user32 = mock_user32

        toolbar_manager.cleanup_resources()

        # Should have destroyed all custom icons
        assert mock_user32.DestroyIcon.call_count == 3
        assert len(toolbar_manager.button_icons["custom_icons"]) == 0

    def test_handle_button_click_with_callback(self, mock_modules, mock_window):
        """Test handling button click with user callback."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        callback_called = False

        def test_callback(button_id, is_playing):
            nonlocal callback_called
            callback_called = True
            assert button_id == WindowsConstants.THUMB_BUTTON_FORWARD
            assert is_playing is None

        toolbar_manager.button_callbacks = {
            WindowsConstants.THUMB_BUTTON_FORWARD: test_callback
        }

        result = toolbar_manager.handle_button_click(
            WindowsConstants.THUMB_BUTTON_FORWARD
        )

        assert result is True
        assert callback_called is True

    def test_handle_button_click_play_pause_auto_toggle(
        self, mock_modules, mock_window
    ):
        """Test play/pause button auto-toggle functionality."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)
        toolbar_manager.is_playing = False

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        toolbar_manager.buttons = [mock_button]
        toolbar_manager._icons = {"play": 11111, "pause": 22222}

        with patch.object(toolbar_manager, "update_play_pause_button") as mock_update:
            mock_update.return_value = True

            result = toolbar_manager.handle_button_click(
                WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
            )

            assert result is True
            mock_update.assert_called_once()


class TestTaskbarManagerWindowMessageHandling:
    """Test TaskbarManager window message handling for media controls."""

    def test_wm_command_button_click_handling(self, mock_modules, mock_window):
        """Test handling WM_COMMAND messages for button clicks."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345

        # Set up toolbar manager with mock
        mock_toolbar = MagicMock()
        mock_toolbar.handle_button_click.return_value = True
        manager.toolbar_manager = mock_toolbar

        # Mock window procedure setup
        from taskbar_manager import WindowsAPISetup

        LRESULT, LONG_PTR, WNDPROCTYPE = WindowsAPISetup.setup_window_proc_types()

        # Simulate WM_COMMAND message with button ID
        wparam = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE  # Button ID in lower word
        result = manager._handle_window_message(
            12345,
            WindowsConstants.WM_COMMAND,
            wparam,
            0,
            None,  # TaskbarButtonCreated
            LONG_PTR,
        )

        assert result == 0  # Message handled
        mock_toolbar.handle_button_click.assert_called_once_with(
            WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        )

    def test_wm_command_non_button_id_ignored(self, mock_modules, mock_window):
        """Test that WM_COMMAND messages with non-button IDs are ignored."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345
        manager.old_wndproc = 99999  # Mock old window procedure

        mock_toolbar = MagicMock()
        manager.toolbar_manager = mock_toolbar

        # Mock user32.CallWindowProcW
        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.CallWindowProcW.return_value = 123
        mock_ctypes.windll.user32 = mock_user32

        from taskbar_manager import WindowsAPISetup

        LRESULT, LONG_PTR, WNDPROCTYPE = WindowsAPISetup.setup_window_proc_types()

        # Simulate WM_COMMAND with non-button ID
        wparam = 9999  # Not a media control button ID
        result = manager._handle_window_message(
            12345, WindowsConstants.WM_COMMAND, wparam, 0, None, LONG_PTR
        )

        # Should call original window procedure
        assert result == 123
        mock_user32.CallWindowProcW.assert_called_once()
        mock_toolbar.handle_button_click.assert_not_called()

    def test_wm_command_no_toolbar_manager(self, mock_modules, mock_window):
        """Test WM_COMMAND handling when toolbar manager is not initialized."""
        from taskbar_manager import TaskbarManager, WindowsConstants

        manager = TaskbarManager(mock_window)
        manager.hwnd = 12345
        manager.old_wndproc = 99999
        manager.toolbar_manager = None

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.CallWindowProcW.return_value = 456
        mock_ctypes.windll.user32 = mock_user32

        from taskbar_manager import WindowsAPISetup

        LRESULT, LONG_PTR, WNDPROCTYPE = WindowsAPISetup.setup_window_proc_types()

        wparam = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        result = manager._handle_window_message(
            12345, WindowsConstants.WM_COMMAND, wparam, 0, None, LONG_PTR
        )

        # Should call original window procedure since no toolbar manager
        assert result == 456
        mock_user32.CallWindowProcW.assert_called_once()


class TestThumbnailToolbarManagerIconConversion:
    """Test icon conversion and loading edge cases."""

    def test_bitmap_to_icon_conversion(self, mock_modules, mock_window):
        """Test bitmap to icon conversion."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_gdi32 = MagicMock()

        # Mock successful conversion
        mock_user32.GetDC.return_value = 111
        mock_user32.ReleaseDC.return_value = 1
        mock_user32.CreateIconIndirect.return_value = 77777
        mock_gdi32.CreateCompatibleBitmap.return_value = 222
        mock_gdi32.GetObjectW.return_value = 24
        mock_gdi32.DeleteObject.return_value = True

        mock_ctypes.windll.user32 = mock_user32
        mock_ctypes.windll.gdi32 = mock_gdi32

        # Mock ctypes functions
        mock_ctypes.create_string_buffer.return_value = b"\x00" * 24

        result = toolbar_manager._bitmap_to_icon(12345)

        assert result == 77777
        mock_user32.CreateIconIndirect.assert_called_once()
        mock_gdi32.DeleteObject.assert_called_once_with(222)  # Mask bitmap cleanup

    def test_bitmap_to_icon_conversion_failure(self, mock_modules, mock_window):
        """Test bitmap to icon conversion failure."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_gdi32 = MagicMock()

        # Mock failed mask bitmap creation
        mock_user32.GetDC.return_value = 111
        mock_gdi32.CreateCompatibleBitmap.return_value = None  # Failed

        mock_ctypes.windll.user32 = mock_user32
        mock_ctypes.windll.gdi32 = mock_gdi32

        result = toolbar_manager._bitmap_to_icon(12345)

        assert result is None

    def test_load_system_icon_extraction_failure(self, mock_modules, mock_window):
        """Test system icon loading failure."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_shell32 = MagicMock()
        mock_shell32.ExtractIconW.return_value = 1  # Indicates failure
        mock_ctypes.windll.shell32 = mock_shell32

        result = toolbar_manager._load_system_icon("shell32.dll", 999)

        assert result is None

    def test_load_system_icon_exception(self, mock_modules, mock_window):
        """Test system icon loading with exception."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_shell32 = MagicMock()
        mock_shell32.ExtractIconW.side_effect = Exception("Access denied")
        mock_ctypes.windll.shell32 = mock_shell32

        result = toolbar_manager._load_system_icon("shell32.dll", 16)

        assert result is None

    @patch("os.path.exists", return_value=True)
    def test_load_icon_from_file_load_failure(
        self, mock_exists, mock_modules, mock_window
    ):
        """Test icon loading failure from existing file."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.LoadImageW.return_value = 0  # Failed to load
        mock_ctypes.windll.user32 = mock_user32

        result = toolbar_manager._load_icon_from_file("test.ico")

        assert result is None

    def test_load_icon_from_file_png_bitmap_failure(self, mock_modules, mock_window):
        """Test PNG loading when bitmap loading fails."""
        from taskbar_manager import ThumbnailToolbarManager, TaskbarManager

        taskbar_manager = TaskbarManager(mock_window)
        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_user32.LoadImageW.return_value = 0  # Failed to load bitmap
        mock_ctypes.windll.user32 = mock_user32

        with patch("os.path.exists", return_value=True):
            result = toolbar_manager._load_icon_from_file("test.png")

        assert result is None

    def test_update_button_icon_with_cleanup(self, mock_modules, mock_window):
        """Test updating button icon with proper cleanup of old icon."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up existing custom icons
        old_icon = 11111
        toolbar_manager.button_icons["custom_icons"] = [old_icon, 22222]

        # Set up a mock button
        from taskbar_manager import THUMBBUTTON

        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        mock_button.hIcon = old_icon
        toolbar_manager.buttons = [mock_button]

        mock_ctypes = mock_modules
        mock_user32 = MagicMock()
        mock_ctypes.windll.user32 = mock_user32

        with patch.object(toolbar_manager, "_load_custom_icon") as mock_load:
            mock_load.return_value = 99999  # New icon

            result = toolbar_manager.update_button_icon(
                WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, "new_icon.ico"
            )

            assert result is True
            assert mock_button.hIcon == 99999

            # Old icon should be destroyed and removed from list
            mock_user32.DestroyIcon.assert_called_once_with(old_icon)
            assert old_icon not in toolbar_manager.button_icons["custom_icons"]
            assert 99999 in toolbar_manager.button_icons["custom_icons"]

    def test_update_button_icon_api_failure(self, mock_modules, mock_window):
        """Test button icon update when API call fails."""
        from taskbar_manager import (
            ThumbnailToolbarManager,
            TaskbarManager,
            WindowsConstants,
        )

        taskbar_manager = TaskbarManager(mock_window)
        taskbar_manager.taskbar_interface = MagicMock()
        taskbar_manager.vtbl = MagicMock()
        taskbar_manager.vtbl.ThumbBarUpdateButtons.return_value = 0x80004005  # E_FAIL
        taskbar_manager.hwnd = 12345

        toolbar_manager = ThumbnailToolbarManager(taskbar_manager)

        # Set up a mock button (don't use spec since THUMBBUTTON is mocked)
        mock_button = MagicMock()
        mock_button.iId = WindowsConstants.THUMB_BUTTON_PLAY_PAUSE
        mock_button.hIcon = 11111
        toolbar_manager.buttons = [mock_button]

        with patch.object(toolbar_manager, "_load_custom_icon") as mock_load:
            mock_load.return_value = 99999  # New icon loads successfully

            result = toolbar_manager.update_button_icon(
                WindowsConstants.THUMB_BUTTON_PLAY_PAUSE, "new_icon.ico"
            )

            assert result is False
            # Icon should still be updated in button structure even if API fails
            assert mock_button.hIcon == 99999


class TestTaskbarManagerIntegration:
    """Integration test for TaskbarManager."""

    def test_complete_thumbnail_workflow(self, mock_modules, mock_window):
        """Test complete thumbnail workflow with parameter verification."""
        from taskbar_manager import TaskbarManager

        manager = TaskbarManager(mock_window)
        manager.hwnd = 99999
        manager.taskbar_button_created = True
        manager.bitmap_creator = MagicMock()

        # Set up comprehensive mocking for the entire workflow
        manager.bitmap_creator.create_thumbnail_bitmap.return_value = (
            4001  # Mock bitmap handle
        )

        with patch("ctypes.windll.dwmapi.DwmSetIconicThumbnail") as mock_set_thumbnail:
            mock_set_thumbnail.return_value = 0

            with patch("ctypes.windll.gdi32.DeleteObject") as mock_delete:
                mock_delete.return_value = True

                # 1. Set clipping rectangle
                manager.SetThumbnailClip(50, 75, 400, 300)

                # 2. Simulate thumbnail request
                manager._delegate_thumbnail_request(99999, 250, 200)

                # 3. Process the request
                manager._handle_thumbnail_request_safe(99999, 250, 200)

                # Verify the complete workflow
                # Clipping should be set correctly
                assert manager.clip_rect.left == 50
                assert manager.clip_rect.top == 75
                assert manager.clip_rect.right == 450  # 50 + 400
                assert manager.clip_rect.bottom == 375  # 75 + 300

                # Request should be delegated correctly
                assert manager.thumbnail_width == 250
                assert manager.thumbnail_height == 200

                # Bitmap should be created with clipped parameters
                manager.bitmap_creator.create_thumbnail_bitmap.assert_called_once_with(
                    50,  # clip_rect.left
                    75,  # clip_rect.top
                    400,  # width
                    300,  # height
                    250,  # thumb_width
                    200,  # thumb_height
                    clip_rect=manager.clip_rect,
                )

                # DWM should be called with correct parameters
                mock_set_thumbnail.assert_called_once_with(99999, 4001, 0)

                # Bitmap should be cleaned up
                mock_delete.assert_called_once_with(4001)



import sys
import os
import pytest

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../nebulatk"))
)

# Now import your package/module
import nebulatk as ntk


@pytest.fixture
def canvas() -> ntk.Window:
    """
    Create a test window for widget testing.

    Returns:
        ntk.Window: A window instance for testing widgets.
    """
    window = ntk.Window(title="Test Window", width=800, height=500)
    yield window
    window.close()


def test_children(canvas: ntk.Window) -> None:
    """
    Test widget children management in the window.

    Args:
        canvas: The test window fixture.
    """
    button = ntk.Button(canvas, text="Button").place()
    assert len(canvas.children) == 1
    button.destroy()
    assert len(canvas.children) == 0


def test_button_properties(canvas: ntk.Window) -> None:
    """
    Test Button widget properties and functionality.

    Args:
        canvas: The test window fixture.
    """
    # Test button creation with various properties
    button = ntk.Button(
        canvas,
        text="Test Button",
        width=100,
        height=30,
        fill="blue",
        text_color="white",
    ).place()

    assert button.text == "Test Button"
    assert button.width == 100
    assert button.height == 30
    assert button.fill == "#0000ffff"
    assert button.text_color == "#ffffff"

    # Test button state changes
    assert not button.state  # Default state should be False
    button.state = True
    assert button.state


def test_label_properties(canvas: ntk.Window) -> None:
    """
    Test Label widget properties and functionality.

    Args:
        canvas: The test window fixture.
    """
    label = ntk.Label(
        canvas,
        text="Test Label",
        width=150,
        height=40,
        text_color="red1",
        fill="yellow1",
    ).place()

    assert label.text == "Test Label"
    assert label.width == 150
    assert label.height == 40
    assert label.text_color == "#ff0000"
    assert label.fill == "#ffff00ff"


def test_slider_functionality(canvas: ntk.Window) -> None:
    """
    Test Slider widget functionality.

    Args:
        canvas: The test window fixture.
    """
    slider = ntk.Slider(
        canvas, width=200, height=30, slider_width=20, slider_height=30
    ).place()

    # Test initial position
    initial_x = slider.button.x

    # Simulate dragging by calling the internal _dragging method
    slider._dragging(initial_x + 50, slider.button.y)

    # Button should have moved
    assert slider.button.x > initial_x

    # Test clamping at boundaries
    # Drag beyond right edge
    slider._dragging(slider.width + 100, slider.button.y)
    assert slider.button.x <= slider.width - slider.button.width

    # Drag beyond left edge
    slider._dragging(-100, slider.button.y)
    assert slider.button.x >= 0



from .button import Button
from .label import Label
from .entry import Entry
from .frame import Frame
from .slider import Slider
from .container import Container

__all__ = ["Button", "Label", "Entry", "Frame", "Slider", "Container"]
